# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser
from math import asin, pi, sqrt

import numpy as np
from matplotlib import pyplot as plt
from termcolor import cprint

from radtools.constants import TODEGREES
from radtools.io.internal import load_template
from radtools.io.tb2j import load_tb2j_model


def rot_angle(x, y, dummy=False):
    r"""
    Rotational angle from 2D vector.

    Mathematically positive => counterclockwise.
    From [0 to 360)

    Parameters
    ----------
    x : float or int
        x coordinate of a vector.
    y : float or int
        y coordinate of a vector.
    """
    # rot_cos = x / (x ** 2 + y ** 2) ** 0.5
    # rot_angle = m.acos(rot_cos) / m.pi * 180
    try:
        sin = abs(y) / sqrt(x**2 + y**2)
    except ZeroDivisionError:
        raise ValueError("Angle is ill defined (x = y = 0).")
    if x > 0:
        if y > 0:
            return asin(sin) * TODEGREES
        elif y == 0:
            return 0
        elif y < 0:
            if not dummy:
                return -asin(sin) * TODEGREES
            return 360 - asin(sin) * TODEGREES
    elif x == 0:
        if y > 0:
            return 90
        elif y == 0:
            raise ValueError("Angle is ill defined (x = y = 0).")
        elif y < 0:
            if not dummy:
                return 90
            return 270
    elif x < 0:
        if y > 0:
            if not dummy:
                return -asin(sin) * TODEGREES
            return 180 - asin(sin) * TODEGREES
        elif y == 0:
            if not dummy:
                return 0
            return 180
        elif y < 0:
            if not dummy:
                return asin(sin) * TODEGREES
            return 180 + asin(sin) * TODEGREES


def atom_mark_to_latex(mark):
    r"""
    Latexifier for atom marks.

    Cr12 -> Cr\ :sub:`12`\.

    Parameters
    ----------
    mark : str
        Mark of atom.

    Returns
    -------
    new_mark : str
        Latex version of the mark.
    """
    numbers = "0123456789"
    new_mark = "$"
    insert_underline = False
    for symbol in mark:
        if symbol in numbers and not insert_underline:
            insert_underline = True
            new_mark += "_{"
        new_mark += symbol
    new_mark += "}$"
    return new_mark


def manager(
    input_filename,
    output_name="exchange",
    what_to_plot="iso",
    draw_cells=False,
    min_distance=None,
    max_distance=None,
    distance=None,
    template_file=None,
    R_vector=None,
    scale_atoms=1,
    scale_data=1,
    title=None,
    form_model=False,
    verbose=False,
):
    r"""
    :ref:`rad-plot-tb2j` script.

    Full documentation on the behaviour is available in the
    :ref:`User Guide <rad-plot-tb2j>`.
    Parameters of the function directly
    correspond to the arguments of the script.

    Parameters
    ----------
    input_filename : str
        Relative or absolute path to the "exchange.out" file, including name and extension of the file.

        Console argument: ``-if`` / ``--input-filename``

        Metavar: "filename"
    output_name : str, default "exchange"
        Seedname for the output files.

        Output files have the following name structure:
        "output-name.display-data-type.png"

        See also: :ref:`example <output-notes>`.

        Console argument: ``-on`` / ``--output-name``

        Metavar: "filename"
    what_to_plot : str, default "iso"
        Type of data for display.

        Specifying the data to be displayed in the graphs.
        Everything is displayed by default, each value in a separate picture.
        Currently available for display: Isotropic exchange parameter, distance, \|DMI\|.

        Console argument: ``-wtp`` / ``--what-to-plot``

        Choices: "all", "iso", "distance", "dmi"
    draw_cells : bool, default False
        Whether to draw the cells.

        If specified then the shapes of all cells
        presented in the model (after filtering) are drawn. (0, 0, 0) is red.

        Console argument: ``-dc`` / ``--draw-cells``
    min_distance : float, default None
        (>=) Minimum distance.

        All the bonds with the distance between atom 1 and atom 2
        lower than minimum distance are excluded from the model.

        Console argument: ``-mind`` / ``--min-distance``

        Metavar: "distance"
    max_distance : float, default None
        (<=) Maximum distance.

        All the bonds with the distance between atom 1 and atom 2
        greater than maximum distance are excluded from the model.

        Console argument: ``-maxd`` / ``--max-distance``

        Metavar: "distance"
    distance : float, default None
        (=) Exact distance.

        Only the bonds with the exact distance remains in the model.
        There is no point in specifying maximum or minimum distance when
        this parameter is provided.

        Console argument: ``-d`` / ``--distance``

        Metavar: "distance"
    template_file : str, default None
        Relative or absolute path to the template file, including the name and extension of the file.

        Console argument: ``-tf`` / ``--template-file``

        Metavar: "filename"
    R_vector : list of int, optional
        R vectors for filtering the spin Hamiltonian.

        In TB2J outputs the bond is defined by atom 1 (from) and atom 2 (to).
        Atom 1 is always located in (0, 0, 0) unit cell, while atom 2 is located in
        R = (i, j, k) unit cell. This parameter tells the script to keep only the
        bonds for which atom 2 is located in one of specified R supercells.
        Supercells are specified by a set of integers separated by spaces.
        They are grouped by three starting from the left and forms a set
        of R vectors. If the last group contains 1 or 2 integers they are ignored.

        Console argument: ``-R`` / ``--R-vector``

        Metavar: "i1 j1 k1 i2 j2 k2 ..."
    scale_atoms : float, default 1
        Scale for the size of atom marks.

        Use it if you want to make atom marks bigger (>1) or smaller (<1).
        Has to be positive.

        Console argument: ``-sa`` / ``--scale-atoms``
    scale_data : float, default 1
        Scale for the size of data text.

        Use it if you want to make data text marks bigger (>1) or smaller (<1).
        Has to be positive.

        Console argument: ``-sd`` / ``--scale-data``
    title : str, optional
        Title for the plots.

        Title is displayed in the picture.

        Console argument: ``-t`` / ``--title``
    form_model : bool, default False
        Force the spin Hamiltonian to have the symmetry of the template.

        Console argument: ``-fm`` / ``--form-model``
    verbose : bool, default False
        Verbose output, propagates to the called methods.

        Console argument: ``-v`` / ``--verbose``
    """

    head, _ = os.path.split(input_filename)
    out_head, out_tail = os.path.split(output_name)
    if len(out_head) == 0:
        out_head = head
    if len(out_tail) == 0:
        out_tail = "exchange"

    output_name = os.path.join(out_head, out_tail)

    # Create the output directory if it does not exist
    if out_head != "":
        os.makedirs(out_head, exist_ok=True)

    # Check input parameters for consistency
    if form_model and template_file is None:
        raise ValueError("--form-model option requires a template file.")

    if distance is not None:
        min_distance = distance
        max_distance = distance

    # Translate sequence of numbers to R vectors
    if R_vector is not None:
        R_vector = np.array(R_vector[: len(R_vector) // 3 * 3], dtype=int).reshape(
            (len(R_vector) // 3, 3)
        )
        R_vector = list(map(tuple, R_vector.tolist()))

    # Read the model
    model = load_tb2j_model(input_filename, quiet=not verbose)

    # Force symmetry of the template
    if template_file is not None:
        template = load_template(template_file)
    else:
        template = None
    if form_model:
        model.form_model(template=template)

    # Filter the model
    model.filter(
        min_distance=min_distance,
        max_distance=max_distance,
        R_vector=R_vector,
        template=template,
    )

    dummy = True
    ha = "center"
    x_min, y_min, z_min, x_max, y_max, z_max = model.space_dimensions
    X = max(abs(x_min), abs(x_max))
    Y = max(abs(y_min), abs(y_max))
    if X == 0 and Y == 0:
        X = Y = 1
    linewidth = 1
    fontsize = 11
    if what_to_plot == "all":
        wtps = ["iso", "distance", "dmi"]
    else:
        wtps = [what_to_plot]
    for wtp in wtps:
        if X < Y:
            fig = plt.figure(figsize=(6.4 * X / Y, 4.8))
        else:
            fig = plt.figure(figsize=(6.4, 4.8 * Y / X))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.set_aspect("equal")
        ax.set_xlabel("x, Angstroms")
        ax.set_ylabel("y, Angstroms")

        for atom1, atom2, R, J in model:
            dis = model.get_distance(atom1, atom2, R)
            x1, y1, z1 = model.get_atom_coordinates(atom1, relative=False)
            x2, y2, z2 = model.get_atom_coordinates(atom2, R, relative=False)
            xm = (x1 + x2) / 2
            ym = (y1 + y2) / 2
            zm = (z1 + z2) / 2

            ax.scatter(x1, y1, s=100 * fig.dpi / 72.0, c="white")
            ax.scatter(x2, y2, s=100 * fig.dpi / 72.0, c="white")

            ax.text(
                x2,
                y2,
                atom_mark_to_latex(atom2.name),
                va="center",
                ha="center",
                fontsize=1.5 * fontsize * scale_atoms,
            )
            ax.text(
                x1,
                y1,
                atom_mark_to_latex(atom1.name),
                va="center",
                ha="center",
                fontsize=1.5 * fontsize * scale_atoms,
                color="#A04F4D",
            )
            if wtp == "iso":
                ax.text(
                    xm,
                    ym,
                    str(round(J.iso, 4)),
                    va="bottom",
                    ha=ha,
                    rotation_mode="anchor",
                    rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                    fontsize=fontsize * scale_data,
                )
            elif wtp == "distance":
                ax.text(
                    xm,
                    ym,
                    str(round(dis, 4)),
                    va="bottom",
                    ha=ha,
                    rotation_mode="anchor",
                    rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                    fontsize=fontsize * scale_data,
                )
            elif wtp == "dmi":
                ax.text(
                    xm,
                    ym,
                    str(round(sqrt(np.sum(J.dmi**2)), 4)),
                    va="bottom",
                    ha=ha,
                    rotation_mode="anchor",
                    rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                    fontsize=fontsize * scale_data,
                )

        xlims = (ax.get_xlim()[0] - 0.5, ax.get_xlim()[1] + 0.5)
        ylims = (ax.get_ylim()[0] - 0.5, ax.get_ylim()[1] + 0.5)

        if draw_cells:
            cells = model.cell_list
            a_x, a_y, a_z = tuple(model.cell[0])
            b_x, b_y, b_z = tuple(model.cell[1])
            Rx_min = 0
            Rx_max = 0
            Ry_min = 0
            Ry_max = 0
            Rz_min = 0
            Rz_max = 0
            for Rx, Ry, Rz in cells:
                Rx_min = min(Rx, Rx_min)
                Rx_max = max(Rx, Rx_max)
                Ry_min = min(Ry, Ry_min)
                Ry_max = max(Ry, Ry_max)
                Rz_min = min(Rz, Rz_min)
                Rz_max = max(Rz, Rz_max)

            for i in range(Rx_min, Rx_max + 2):
                for j in range(Ry_min, Ry_max + 2):
                    for k in range(Rz_min, Rz_max + 2):
                        ax.plot(
                            [Rx_min * a_x + j * b_x, (Rx_max + 1) * a_x + j * b_x],
                            [Rx_min * a_y + j * b_y, (Rx_max + 1) * a_y + j * b_y],
                            linewidth=linewidth,
                            color="#DFDFDF",
                        )
                        ax.plot(
                            [i * a_x + Ry_min * b_x, i * a_x + (Ry_max + 1) * b_x],
                            [i * a_y + Ry_min * b_y, i * a_y + (Ry_max + 1) * b_y],
                            linewidth=linewidth,
                            color="#DFDFDF",
                        )
            ax.plot(
                np.array([0, a_x, a_x + b_x, b_x, 0]),
                np.array([0, a_y, a_y + b_y, b_y, 0]),
                linewidth=linewidth,
                color="#CA7371",
            )

        ax.set_xlim(*xlims)
        ax.set_ylim(*ylims)

        if title is not None:
            ax.set_title(title, fontsize=1.5 * fontsize)

        png_path = f"{output_name}.{wtp}.png"
        plt.savefig(png_path, dpi=400, bbox_inches="tight")
        cprint(f"2D plot with {wtp} is in {os.path.abspath(png_path)}", "blue")


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-if",
        "--input-filename",
        required=True,
        metavar="filename",
        type=str,
        help='Relative or absolute path to the "exchange.out" file, including name and extension of the file.',
    )
    parser.add_argument(
        "-on",
        "--output-name",
        default="exchange",
        metavar="filename",
        type=str,
        help="Seedname for the output files.",
    )
    parser.add_argument(
        "-wtp",
        "--what-to-plot",
        default="iso",
        type=str,
        choices=[
            "all",
            "iso",
            "distance",
            "dmi",
        ],
        help="Type of data for display.",
    )
    parser.add_argument(
        "-dc",
        "--draw-cells",
        default=False,
        action="store_true",
        help="Whether to draw the cells.",
    )
    parser.add_argument(
        "-mind",
        "--min-distance",
        default=None,
        metavar="distance",
        type=float,
        help="(>=) Minimum distance.",
    )
    parser.add_argument(
        "-maxd",
        "--max-distance",
        default=None,
        metavar="distance",
        type=float,
        help="(<=) Maximum distance.",
    )
    parser.add_argument(
        "-d",
        "--distance",
        default=None,
        metavar="distance",
        type=float,
        help="(=) Exact distance.",
    )
    parser.add_argument(
        "-tf",
        "--template-file",
        default=None,
        metavar="filename",
        type=str,
        help="Relative or absolute path to the template file, including the name and extension of the file.",
    )
    parser.add_argument(
        "-R",
        "--R-vector",
        default=None,
        metavar="i1 j1 k1 i2 j2 k2 ...",
        type=int,
        nargs="*",
        help="R vectors for filtering the spin Hamiltonian.",
    )
    parser.add_argument(
        "-sa",
        "--scale-atoms",
        default=1,
        type=float,
        help="Scale for the size of atom marks.",
    )
    parser.add_argument(
        "-sd",
        "--scale-data",
        default=1,
        type=float,
        help="Scale for the size of data text.",
    )
    parser.add_argument(
        "-t",
        "--title",
        default=None,
        type=str,
        help="Title for the plots.",
    )
    parser.add_argument(
        "-fm",
        "--form-model",
        default=False,
        action="store_true",
        help="Force the spin Hamiltonian to have the symmetry of the template.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Verbose output, propagates to the called methods.",
    )

    return parser



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser

import numpy as np
from termcolor import cprint


def manager(input_filename, span=0.1, output_name=""):
    r"""
    :ref:`rad-identify-wannier-centres` script.

    Full documentation on the behaviour is available in the
    :ref:`User Guide <rad-identify-wannier-centres>`.
    Parameters of the function directly
    correspond to the arguments of the script.

    Parameters
    ----------
    input_filename : str
        Relative or absolute path to the "_centres.xyz" file

        Identified Wannier centres are stored in the "filename_identified" file.

        Console argument: ``-if`` / ``--input-filename``

        Metavar: "filename"

        .. versionchanged:: 0.8.0 Renamed from ``filename``
    span : float, default 0.1
        Distance tolerance between centre and atom. (in Angstroms)

        If some centres remain unidentified try to increase the span.

        Console argument: ``-s`` / ``--span``
    output_name : str, default ""
        Seedname for the output files.

        See also: :ref:`example <output-notes>`

        Console argument: ``-on`` / ``--output-name``
    """

    # Get the output path and the output name from the input filename
    # if they are not specified
    head, tail = os.path.split(input_filename)
    out_head, out_tail = os.path.split(output_name)
    if len(out_head) == 0:
        out_head = head
    if len(out_tail) == 0:
        out_tail = tail + "_identified"

    # Create the output directory if it does not exist
    if out_head != "":
        os.makedirs(out_head, exist_ok=True)

    output_name = os.path.join(out_head, out_tail)

    # Set the default separation tolerance for the span
    separation_tolerance = 10e-8

    # Read atoms and centres
    atom_counter = {}
    with open(input_filename, "r") as file:
        # Read the number of atoms and centres
        n = int(file.readline())

        # Read the file stats
        file_stats = file.readline()

        # Read the atoms and centres
        centres = []
        atoms = []
        for _ in range(0, n):
            line = file.readline().split()
            if line[0] == "X":
                centres.append([f"X", np.array(list(map(float, line[1:])))])
            else:
                if line[0] in atom_counter:
                    atom_counter[line[0]] += 1
                else:
                    atom_counter[line[0]] = 1
                atoms.append(
                    [
                        f"{line[0]}{atom_counter[line[0]]}",
                        np.array(list(map(float, line[1:]))),
                    ]
                )

    # Identify the centres by the closest atom
    for centre in centres:
        min_span = 10000
        name = "None"

        # Find the closest atom
        for atom, a_coord in atoms:
            if np.linalg.norm((centre[1] - a_coord)) < min_span:
                min_span = np.linalg.norm((centre[1] - a_coord))
                name = atom

        if min_span - span > separation_tolerance:
            cprint(
                f"Centre {centre} unidentified, " + "try to increase span",
                "yellow",
            )
            print(
                f"    span limit = {span}\n"
                + f"    minimum distance to the atom ({name}) = {min_span:.8f}\n"
            )
            centre[0] = f"None, closest: {name} ({min_span:.8f})"
        else:
            centre[0] = name

    # Save the count of centres in atoms
    count = {}

    # Write the output
    with open(output_name, "w", encoding="utf-8") as file:
        file.write(f"{len(atoms) + len(centres):6.0f}\n")
        file.write(f"{file_stats}")
        for centre, coordinate in centres:
            if centre in count:
                count[centre] += 1
            else:
                count[centre] = 1
            file.write(
                f"X      "
                + f"{coordinate[0]:14.8f}   "
                + f"{coordinate[1]:14.8f}   "
                + f"{coordinate[2]:14.8f}"
                + f"   ->   {centre}\n"
            )
        for atom, coordinate in atoms:
            file.write(
                f"{atom.translate(str.maketrans('','','0123456789')):2}     "
                + f"{coordinate[0]:14.8f}   "
                + f"{coordinate[1]:14.8f}   "
                + f"{coordinate[2]:14.8f}"
                + f"   ->   {atom}\n"
            )
    cprint(f"Results are in {os.path.abspath(output_name)}", "blue")
    count = "\n".join(
        [f"{centre:6} -> {amount:4} WF" for centre, amount in count.items()]
    )
    print(f"Found:\n{count}")


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-if",
        "--input-filename",
        required=True,
        metavar="filename",
        type=str,
        help='Relative or absolute path to the "_centres.xyz" file',
    )
    parser.add_argument(
        "-s",
        "--span",
        default=0.1,
        type=float,
        help="Distance tolerance between centre and atom. (in Angstroms)",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        default="",
        type=str,
        help="Seedname for the output files.",
    )

    return parser



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np
from termcolor import cprint

from radtools.decorate.array import print_2d_array
from radtools.decorate.axes import plot_hlines
from radtools.decorate.stats import logo
from radtools.io.internal import load_template
from radtools.io.tb2j import load_tb2j_model
from radtools.magnons.dispersion import MagnonDispersion
from radtools.spinham.constants import TXT_FLAGS


def manager(
    input_filename,
    spin,
    template_file=None,
    output_name="magnon_dispersion",
    spiral_vector=None,
    rotation_axis=None,
    k_points=None,
    k_path=None,
    form_model=False,
    R_vector=None,
    max_distance=None,
    min_distance=None,
    save_txt=False,
    interactive=False,
    verbose=False,
    bravais_type=None,
    join_output=False,
    nodmi=False,
    no_anisotropic=False,
):
    r"""
    :ref:`rad-plot-tb2j-magnons` script.

    Full documentation on the behaviour is available in the
    :ref:`User Guide <rad-plot-tb2j-magnons>`.
    Parameters of the function directly
    correspond to the arguments of the script.

    Parameters
    ----------
    input_filename : str
        Relative or absolute path to the "exchange.out" file, including the name and extension of the file itself.

        Console argument: ``-if`` / ``--input-filename``

        Metavar: "filename"
    spin : list of str, optional
        Spin of the atoms in the model.

        For each atom, which has at least one bond connected to it is necessary to specify
        spin vector. The spin vector is specified in the form of atom's name followed by
        three numbers, separated by spaces.
        The numbers represent the x, y, and z components of the spin vector.

        Console argument: ``-s`` / ``--spin``

        Metavar: "Atom S_x S_y S_z"
    template_file : str, optional
        Relative or absolute path to the template file, including the name and extension of the file.

        Console argument: ``-tf`` / ``--template-file``

        Metavar: "filename"
    output_name : str, default "magnon_dispersion"
        Seedname for the output files.

        Console argument: ``-on`` / ``--output-name``

        Metavar: "filename"
    spiral_vector : list of 3 float, optional
        Spin spiral vector. Relative to the reciprocal cell.

        Console argument: ``-Q`` / ``--spiral-vector``

        Metavar: ("Q_x", "Q_y", "Q_z")
    rotation_axis : list of 3 float, optional
        Direction of global rotation axis. In absolute coordinates in real space.

        Console argument: ``-ra`` / ``--rotation-axis``

        Metavar: ("n_x", "n_y", "n_z")
    k_points : list of str, optional
        Additional high-symmetry k-points.

        Coordinates are relative to the reciprocal cell.

        .. versionadded:: 0.8.9

        Console argument: ``-kps`` / ``--k-points``

        Metavar: "name label xrel yrel zrel ..."
    k_path : str, optional
        Path in reciprocal space for the magnon dispersion.

        Console argument: ``-kp`` / ``--k-path``

        Metavar: "G-X-M-G|G-Y"
    form_model : bool, default False
        Whether to form the spinham based on the template.

        Console argument: ``-fm`` / ``--form-model``
    R_vector : list of int, optional
        R vectors for filtering the spin Hamiltonian.

        In TB2J outputs the bond is defined by atom 1 (from) and atom 2 (to).
        Atom 1 is always located in (0, 0, 0) unit cell, while atom 2 is located in
        R = (i, j, k) unit cell. This parameter tells the script to keep only the
        bonds for which atom 2 is located in one of specified R supercells.
        Supercells are specified by a set of integers separated by spaces.
        They are grouped by three starting from the left and forms a set
        of R vectors. If the last group contains 1 or 2 integers they are ignored.

        Console argument: ``-R`` / ``--R-vector``

        Metavar: "i1 j1 k1 i2 j2 k2 ..."
    max_distance : float, optional
        (<=) Maximum distance.

        All the bonds with the distance between atom 1 and atom 2
        greater than maximum distance are excluded from the model.

        Console argument: ``-maxd`` / ``--max-distance``

        .. versionadded:: 0.8.0
    min_distance : float, optional
        (>=) Minimum distance.

        All the bonds with the distance between atom 1 and atom 2
        lower than minimum distance are excluded from the Hamiltonian.

        Console argument: ``-mind`` / ``--min-distance``

        .. versionadded:: 0.8.0
    save_txt : bool, default False
        Whether to save data to .txt file.

        Two files appears: "output-name.txt" and "output-name_info.txt".
        First one contains raw data of the graph,
        second one contains information about the parameters.

        Console argument: ``-st`` / ``--save-txt``
    interactive : bool, default False
        Whether to show interactive plot.

        Console argument: ``-i`` / ``--interactive``
    verbose : bool, default False
        Verbose output, propagates to the called methods.

        Console argument: ``-v`` / ``--verbose``
    bravais_type : str, optional
        Bravais lattice type. If not provided, the type is identified automatically.

        It does not force the Bravais lattice type on the model,
        but tries to reach the desired type by reducing the
        numerical accuracy in the :py:func:`lepage` algorithm.

        Console argument: ``-bt`` / ``--bravais-type``

        Choices: "CUB", "FCC", "BCC", "TET", "BCT", "ORC", "ORCF", "ORCI", "ORCC", "HEX", "RHL", "MCL", "MCLC", "TRI"
    join_output : bool, default False
        Whether to join the output files into a single file.

        Console argument: ``-jo`` / ``--join-output``
    nodmi : bool, default False
        Whether to ignore DMI in the spinham.

        Console argument: ``-nodmi``
    no_anisotropic : bool, default False
        Whether to ignore anisotropic symmetric exchange in the spinham.

        Console argument: ``-noa`` / ``--no-anisotropic``
    """

    head, _ = os.path.split(input_filename)
    out_head, out_tail = os.path.split(output_name)
    if len(out_head) == 0:
        out_head = head
    if len(out_tail) == 0:
        out_tail = "magnon_dispersion"

    output_name = os.path.join(out_head, out_tail)

    # Create the output directory if it does not exist
    if out_head != "":
        os.makedirs(out_head, exist_ok=True)

    # Check input parameters for consistency
    if form_model and template_file is None:
        raise ValueError(
            "--form-model option requires a template file (--template-filename)."
        )

    # Translate sequence of numbers to R vectors
    if R_vector is not None:
        R_vector = np.array(R_vector[: len(R_vector) // 3 * 3], dtype=int).reshape(
            (len(R_vector) // 3, 3)
        )
        R_vector = list(map(tuple, R_vector.tolist()))

    # Read the spinham
    spinham = load_tb2j_model(
        input_filename, quiet=not verbose, bravais_type=bravais_type
    )

    cprint(f"{spinham.variation} crystal detected", "green")

    # Get k points of the spinham
    kp = spinham.kpoints

    # Add additional points
    if k_points is not None:
        for i in range(len(k_points) // 5):
            point_name = k_points[5 * i]
            point_label = k_points[5 * i + 1]
            point_coordinates = list(map(float, k_points[5 * i + 2 : 5 * i + 5]))
            kp.add_hs_point(
                name=point_name, coordinates=point_coordinates, label=point_label
            )

    # Set custom k path
    if k_path is not None:
        kp.path = k_path

    if verbose:
        print("Predefined high symmetry k points:")
        for name in kp.hs_names:
            print(f"  {name} : {kp.hs_coordinates[name]}")

    # Force symmetry of the template
    if template_file is not None:
        template = load_template(template_file)
    else:
        template = None

    # Form the spinham based on the template or filter it
    if form_model:
        spinham.form_model(template=template)
    else:
        spinham.filter(
            min_distance=min_distance,
            max_distance=max_distance,
            R_vector=R_vector,
            template=template,
        )

    # Set the spin of the atoms
    if spin is not None:
        for i in range(len(spin) // 4):
            atom_name = spin[4 * i]
            atom = spinham.get_atom(atom_name)
            atom_spin = list(map(float, spin[4 * i + 1 : 4 * i + 4]))
            atom.spin_vector = atom_spin

    # Get the magnon dispersion
    dispersion = MagnonDispersion(
        spinham,
        Q=spiral_vector,
        n=rotation_axis,
        nodmi=nodmi,
        noaniso=no_anisotropic,
    )

    fig, ax = plt.subplots()

    omegas = dispersion(kp)

    ax.set_xticks(kp.coordinates(), kp.labels, fontsize=15)
    ax.set_ylabel("E, meV", fontsize=15)
    ax.vlines(
        kp.coordinates(),
        0,
        1,
        transform=ax.get_xaxis_transform(),
        colors="black",
        alpha=0.5,
        lw=1,
        ls="dashed",
    )
    colors = ["#174FD5", "#F8AB00", "#0CE1A2", "#FF003C", "#46EC00", "#9823C9"]
    i = 0
    for omega in omegas:
        ax.plot(kp.flatten_points(), omega, color=colors[i % len(colors)])
        i += 1

    ax.set_xlim(kp.flatten_points()[0], kp.flatten_points()[-1])
    plot_hlines(ax, [0])

    if save_txt:
        main_separator = "=" * 80 + "\n"
        info = [
            main_separator,
            logo(date_time=True, line_length=80),
            f"\nMagnon dispersion is computed based on the file:\n{os.path.abspath(input_filename)}\n",
        ]
        if template_file is not None:
            info.append(f"With template file:\n{os.path.abspath(template_file)}\n")

        info.append(main_separator)
        info.append(TXT_FLAGS["cell"] + "\n")
        info.append(
            print_2d_array(
                spinham.cell,
                borders=False,
                fmt="^.8f",
                print_result=False,
                header_row=["x", "y", "z"],
            )
            + "\n"
        )
        info.append(f"Detected Bravais lattice: {spinham.variation}\n")
        info.append(main_separator)
        info.append(TXT_FLAGS["atoms"] + "\n")
        header_column = []
        header_row = [
            f"Index Name",
            "a1 (rel)",
            "a2 (rel)",
            "a3 (rel)",
            "S_x",
            "S_y",
            "S_z",
        ]
        atom_data = np.zeros((len(spinham.magnetic_atoms), 6))
        for a_i, atom in enumerate(spinham.magnetic_atoms):
            header_column.append(f"{atom.index:<5} {atom.name:<4}")
            atom_data[a_i, :3] = spinham.get_atom_coordinates(atom, relative=True)
            atom_data[a_i, 3:] = atom.spin_vector
        info.append(
            print_2d_array(
                atom_data,
                borders=False,
                fmt="^.8f",
                print_result=False,
                header_row=header_row,
                header_column=header_column,
            )
            + "\n"
        )

        if spiral_vector is not None:
            info.append(f"Spiral vector: {dispersion.Q} (relative: {spiral_vector})\n")
            info.append(f"Rotation axis: {dispersion.n}\n")

        info.append(main_separator)
        info.append(TXT_FLAGS["kpath"] + "\n")
        info.append(f"  {kp.path_string}\n")
        info.append(TXT_FLAGS["kpoints"] + "\n")
        names_data = []
        header_column = []
        for name in kp.hs_names:
            header_column.append(f"{name:<6}")
            names_data.append(kp.hs_coordinates[name])
        info.append(
            print_2d_array(
                names_data,
                borders=False,
                fmt=">.8f",
                print_result=False,
                header_column=header_column,
            )
            + "\n"
        )

        info.append(TXT_FLAGS["klabels"] + "\n")
        labels_data = []
        header_column = []
        for i in range(len(kp.labels)):
            header_column.append(f"{kp.labels[i]:<10}")
            labels_data.append([kp.coordinates()[i]])
        info.append(
            print_2d_array(
                labels_data,
                borders=False,
                fmt=">.8f",
                print_result=False,
                header_column=header_column,
            )
            + "\n"
        )
        info.append(main_separator)
        info.append(TXT_FLAGS["dispersion"] + "\n")

        header = ["Coordinate"]
        for i in range(omegas.shape[0]):
            header.append(f' {f"omega_{i}":>14}')
        header.append(f"    {'b1 (rel)':>10} {'b2 (rel)':>10} {'b3 (rel)':>10}")
        header.append(f"    {'k_x (abs)':>10} {'k_y (abs)':>10} {'k_z (abs)':>10}")
        info.append("".join(header))

        if not join_output:
            filename = f"{output_name}_data.txt"
            info.extend(
                [
                    "\n",
                    TXT_FLAGS["separate"],
                    "\n",
                    os.path.abspath(filename),
                ]
            )
            info = "".join(info)
            with open(f"{output_name}_info.txt", "w", encoding="utf-8") as file:
                file.write(info)
            info = ""
            comments = "#"
        else:
            info = "".join(info)
            filename = f"{output_name}.txt"
            comments = ""

        np.savetxt(
            filename,
            np.concatenate(
                (
                    [kp.flatten_points()],
                    omegas,
                    kp.points(relative=True).T,
                    kp.points(relative=False).T,
                ),
                axis=0,
            ).T,
            fmt="%.8f"
            + " %.8e" * omegas.shape[0]
            + "   "
            + " %.8f" * 3
            + "   "
            + " %.8f" * 3,
            header=info,
            comments=comments,
        )
    if interactive:
        plt.show()
    else:
        plt.savefig(
            f"{output_name}.png",
            bbox_inches="tight",
            dpi=600,
        )
        cprint(
            f"Results are in {os.path.abspath(out_head)}, seedname: {out_tail}.", "blue"
        )


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-if",
        "--input-filename",
        required=True,
        metavar="filename",
        type=str,
        help='Relative or absolute path to the "exchange.out" file, including the name and extension of the file itself.',
    )
    parser.add_argument(
        "-s",
        "--spin",
        required=True,
        metavar="Atom S_x S_y S_z",
        type=str,
        nargs="*",
        help="Spin of the atoms in the model.",
    )
    parser.add_argument(
        "-tf",
        "--template-file",
        default=None,
        metavar="filename",
        type=str,
        help="Relative or absolute path to the template file, including the name and extension of the file.",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        default="magnon_dispersion",
        metavar="filename",
        type=str,
        help="Seedname for the output files.",
    )
    parser.add_argument(
        "-Q",
        "--spiral-vector",
        default=None,
        metavar=("Q_x", "Q_y", "Q_z"),
        type=float,
        nargs=3,
        help="Spin spiral vector. Relative to the reciprocal cell.",
    )
    parser.add_argument(
        "-ra",
        "--rotation-axis",
        default=None,
        metavar=("n_x", "n_y", "n_z"),
        type=float,
        nargs=3,
        help="Direction of global rotation axis. In absolute coordinates in real space.",
    )
    parser.add_argument(
        "-kps",
        "--k-points",
        default=None,
        metavar="name label xrel yrel zrel ...",
        type=str,
        nargs="*",
        help="Additional high-symmetry k-points.",
    )
    parser.add_argument(
        "-kp",
        "--k-path",
        default=None,
        metavar="G-X-M-G|G-Y",
        type=str,
        help="Path in reciprocal space for the magnon dispersion.",
    )
    parser.add_argument(
        "-fm",
        "--form-model",
        default=False,
        action="store_true",
        help="Whether to form the spinham based on the template.",
    )
    parser.add_argument(
        "-R",
        "--R-vector",
        default=None,
        metavar="i1 j1 k1 i2 j2 k2 ...",
        type=int,
        nargs="*",
        help="R vectors for filtering the spin Hamiltonian.",
    )
    parser.add_argument(
        "-maxd",
        "--max-distance",
        default=None,
        type=float,
        help="(<=) Maximum distance.",
    )
    parser.add_argument(
        "-mind",
        "--min-distance",
        default=None,
        type=float,
        help="(>=) Minimum distance.",
    )
    parser.add_argument(
        "-st",
        "--save-txt",
        default=False,
        action="store_true",
        help="Whether to save data to .txt file.",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        default=False,
        action="store_true",
        help="Whether to show interactive plot.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Verbose output, propagates to the called methods.",
    )
    parser.add_argument(
        "-bt",
        "--bravais-type",
        default=None,
        type=str,
        choices=[
            "CUB",
            "FCC",
            "BCC",
            "TET",
            "BCT",
            "ORC",
            "ORCF",
            "ORCI",
            "ORCC",
            "HEX",
            "RHL",
            "MCL",
            "MCLC",
            "TRI",
        ],
        help="Bravais lattice type. If not provided, the type is identified automatically.",
    )
    parser.add_argument(
        "-jo",
        "--join-output",
        default=False,
        action="store_true",
        help="Whether to join the output files into a single file.",
    )
    parser.add_argument(
        "-nodmi",
        default=False,
        action="store_true",
        help="Whether to ignore DMI in the spinham.",
    )
    parser.add_argument(
        "-noa",
        "--no-anisotropic",
        default=False,
        action="store_true",
        help="Whether to ignore anisotropic symmetric exchange in the spinham.",
    )

    return parser



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from os import makedirs
from os.path import abspath, isfile, join

from termcolor import cprint
from tqdm import tqdm

from radtools.dos.dos import prepare_custom_pdos
from radtools.dos.pdos import PDOS
from radtools.dos.plotting import COLOURS, plot_projected

__all__ = [
    "plot_custom_pdos",
    "plot_orbital_resolved",
    "plot_atom_resolved",
    "plot_atom_to_total",
]


def plot_custom_pdos(
    dos,
    custom,
    output_root=".",
    energy_window=None,
    dos_window=None,
    efermi=0.0,
    relative=False,
    normalize=False,
    interactive=False,
    save_pickle=False,
    save_txt=False,
    background_total=False,
    colours=COLOURS,
    labels=None,
    legend_fontsize=12,
    axes_labels_fontsize=14,
    title_fontsize=18,
):
    cprint("Plotting custom plot", "green")
    print("Input is understood as:")
    projectors = []

    # Check if labels are provided and correct
    if (
        labels is not None
        and len(labels) != len(custom)
        and len(labels) != len(custom) + 1
    ):
        raise ValueError(
            f"Got {len(labels)} labels, but {len(custom)} PDOS, have to be the same or n custom, n+1 labels."
        )

    # Process some of the predefined labels
    if labels is not None and len(labels) == len(custom) + 1:
        if labels[0].lower() == "none":
            total_label = None
        elif labels[0].lower() == "default":
            total_label = "default"
        else:
            total_label = labels[0]
        labels = labels[1:]
    else:
        total_label = "default"

    # Set projectors
    if labels is None:
        projectors = custom
    else:
        projectors = labels

    # Get PDOS array
    pdos = prepare_custom_pdos(dos=dos, custom=custom, quiet=False)

    # Create PDOS object
    if background_total:
        pdos = PDOS(
            energy=dos.energy,
            pdos=pdos,
            ldos=dos.total_pdos(),
            projectors_group="Total PDOS",
            projectors=projectors,
            spin_pol=dos.spin_pol,
        )
    else:
        pdos = PDOS(
            energy=dos.energy,
            pdos=pdos,
            projectors_group="Total (sum)",
            projectors=projectors,
            spin_pol=dos.spin_pol,
        )

    # Compute output_name
    if isfile(join(output_root, "custom.png")):
        i = 1
        while isfile(join(output_root, f"custom{i}.png")):
            i += 1
        output_name = f"custom{i}"
    else:
        output_name = "custom"

    # Save txt
    if save_txt:
        pdos.dump_txt(join(output_root, f"{output_name}.txt"))

    # Plot
    plot_projected(
        pdos=pdos,
        efermi=efermi,
        output_name=join(output_root, output_name),
        xlim=energy_window,
        ylim=dos_window,
        relative=relative,
        normalize=normalize,
        interactive=interactive,
        save_pickle=save_pickle,
        colours=colours,
        total_label=total_label,
        legend_fontsize=legend_fontsize,
        axes_labels_fontsize=axes_labels_fontsize,
        title_fontsize=title_fontsize,
    )
    cprint(f"Result is in {abspath(join(output_root, f'{output_name}.png'))}", "blue")


def plot_orbital_resolved(
    dos,
    output_root=".",
    energy_window=None,
    dos_window=None,
    efermi=0.0,
    separate=False,
    relative=False,
    normalize=False,
    interactive=False,
    save_pickle=False,
    save_txt=False,
    background_total=False,
    colours=COLOURS,
    legend_fontsize=12,
    axes_labels_fontsize=14,
    title_fontsize=18,
):
    cprint("Orbital-resolved PDOS:", "green")
    local_output = join(output_root, "orbital-resolved")
    makedirs(local_output, exist_ok=True)

    data = {}
    for atom, atom_number, wfc, wfc_number in dos:
        if atom not in data:
            data[atom] = []
        data[atom].append((wfc, wfc_number))

    # Avoid repetitions
    for atom in data:
        data[atom] = list(set(data[atom]))

    for atom in data:
        for wfc, wfc_number in data[atom]:
            if separate:
                atom_numbers = dos.atom_numbers(atom)
            else:
                atom_numbers = [None]

            for atom_number in tqdm(atom_numbers, desc=f"  {atom} {wfc} #{wfc_number}"):
                if separate:
                    atom_name = f"{atom}#{atom_number}"
                else:
                    atom_name = atom

                title = f"PDOS for {atom_name} ({wfc} #{wfc_number})"

                pdos = dos.pdos(
                    atom=atom,
                    wfc=wfc,
                    wfc_numbers=wfc_number,
                    atom_numbers=atom_number,
                    background_total=background_total,
                )
                if background_total:
                    pdos.projectors_group = "Total PDOS"
                if save_txt:
                    pdos.dump_txt(
                        join(local_output, f"{atom_name}_{wfc}#{wfc_number}.txt")
                    )
                plot_projected(
                    pdos=pdos,
                    efermi=efermi,
                    output_name=join(local_output, f"{atom_name}_{wfc}#{wfc_number}"),
                    title=title,
                    xlim=energy_window,
                    ylim=dos_window,
                    relative=relative,
                    normalize=normalize,
                    interactive=interactive,
                    save_pickle=save_pickle,
                    colours=colours,
                    legend_fontsize=legend_fontsize,
                    axes_labels_fontsize=axes_labels_fontsize,
                    title_fontsize=title_fontsize,
                )
    cprint(f"Results are in {abspath(local_output)}", "blue")


def plot_atom_resolved(
    dos,
    output_root=".",
    energy_window=None,
    dos_window=None,
    efermi=0.0,
    separate=False,
    relative=False,
    normalize=False,
    interactive=False,
    save_pickle=False,
    save_txt=False,
    background_total=False,
    colours=COLOURS,
    legend_fontsize=12,
    axes_labels_fontsize=14,
    title_fontsize=18,
):
    cprint("Orbital's contribution for each atom.", "green")
    local_output = join(output_root, "atom-resolved")
    makedirs(local_output, exist_ok=True)

    for atom in dos.atoms:
        if separate:
            atom_numbers = dos.atom_numbers(atom)
        else:
            atom_numbers = [None]
        for atom_number in tqdm(atom_numbers, desc=f"  {atom}"):
            if separate:
                atom_name = f"{atom}#{atom_number}"
            else:
                atom_name = atom
            projectors = []
            pdos = []
            for wfc in dos.wfcs(atom, atom_number):
                for wfc_number in dos.wfc_numbers(atom, wfc, atom_number):
                    projectors.append(f"{wfc} #{wfc_number}")
                    pdos.append(dos.pdos(atom, wfc, wfc_number, atom_number).ldos)

            if background_total:
                pdos = PDOS(
                    energy=dos.energy,
                    pdos=pdos,
                    ldos=dos.total_pdos(),
                    projectors_group=atom_name,
                    projectors=projectors,
                    spin_pol=dos.spin_pol,
                )
                pdos.projectors_group = "Total PDOS"
            else:
                pdos = PDOS(
                    energy=dos.energy,
                    pdos=pdos,
                    projectors_group=atom_name,
                    projectors=projectors,
                    spin_pol=dos.spin_pol,
                )
            title = f"PDOS for {atom_name}"
            if save_txt:
                pdos.dump_txt(join(local_output, f"{atom_name}.txt"))
            plot_projected(
                pdos=pdos,
                efermi=efermi,
                output_name=join(local_output, atom_name),
                title=title,
                xlim=energy_window,
                ylim=dos_window,
                relative=relative,
                normalize=normalize,
                interactive=interactive,
                save_pickle=save_pickle,
                colours=colours,
                legend_fontsize=legend_fontsize,
                axes_labels_fontsize=axes_labels_fontsize,
                title_fontsize=title_fontsize,
            )
    cprint(f"Results are in {abspath(local_output)}", "blue")


def plot_atom_to_total(
    dos,
    output_root=".",
    energy_window=None,
    dos_window=None,
    efermi=0.0,
    separate=False,
    relative=False,
    normalize=False,
    interactive=False,
    save_pickle=False,
    save_txt=False,
    background_total=False,
    colours=COLOURS,
    legend_fontsize=12,
    axes_labels_fontsize=14,
    title_fontsize=18,
):
    cprint("Atom's contributions into total PDOS:", "green")
    projectors = []
    pdos = []
    for atom in dos.atoms:
        if separate:
            atom_numbers = dos.atom_numbers(atom)
        else:
            atom_numbers = [None]
        for atom_number in atom_numbers:
            if separate:
                atom_name = f"{atom}#{atom_number}"
            else:
                atom_name = atom
            projectors.append(atom_name)
            for i, wfc in enumerate(dos.wfcs(atom, atom_number)):
                if i == 0:
                    ldos = dos.pdos(atom, wfc, None, atom_number).ldos
                else:
                    ldos += dos.pdos(atom, wfc, None, atom_number).ldos
            pdos.append(ldos)
    if background_total:
        pdos = PDOS(
            energy=dos.energy,
            pdos=pdos,
            ldos=dos.total_pdos(),
            projectors_group="Total PDOS",
            projectors=projectors,
            spin_pol=dos.spin_pol,
        )
        pdos.projectors_group = "Total PDOS"
    else:
        pdos = PDOS(
            energy=dos.energy,
            pdos=pdos,
            projectors_group="Total PDOS",
            projectors=projectors,
            spin_pol=dos.spin_pol,
        )

    title = f"Atom contribution in PDOS"
    if save_txt:
        pdos.dump_txt(join(output_root, "atomic-contributions.txt"))
    plot_projected(
        pdos=pdos,
        efermi=efermi,
        output_name=join(output_root, "atomic-contributions"),
        title=title,
        xlim=energy_window,
        ylim=dos_window,
        relative=relative,
        normalize=normalize,
        interactive=interactive,
        save_pickle=save_pickle,
        colours=colours,
        legend_fontsize=legend_fontsize,
        axes_labels_fontsize=axes_labels_fontsize,
        title_fontsize=title_fontsize,
    )
    cprint(f"Result is in {abspath(output_root)}", "blue")



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Iterable

__all__ = ["plot_hlines", "plot_vlines"]


def plot_hlines(ax, positions, **kwargs):
    r"""
    Plot horizontal lines.

    Parameters
    ----------
    ax : :matplotlib:`axes`
        Axes to plot on.
    positions : list
        List of y positions.
    **kwargs
        Keyword arguments to be passed to :meth:`matplotlib.axes.Axes.hlines`.
    """

    if not isinstance(positions, Iterable):
        positions = [positions]

    if "transfrom" in kwargs:
        kwargs.pop("transform")
    if "color" not in kwargs:
        kwargs["color"] = "grey"
    if "linewidths" not in kwargs:
        kwargs["linewidths"] = 0.5
    if "linestyles" not in kwargs:
        kwargs["linestyles"] = "dashed"

    ax.hlines(positions, 0, 1, transform=ax.get_yaxis_transform(), **kwargs)


def plot_vlines(ax, positions, **kwargs):
    r"""
    Plot vertical lines.

    Parameters
    ----------
    ax : :matplotlib:`axes`
        Axes to plot on.
    positions : list
        List of x positions.
    **kwargs
        Keyword arguments to be passed to :meth:`matplotlib.axes.Axes.vlines`.
    """

    if not isinstance(positions, Iterable):
        positions = [positions]

    if "transfrom" in kwargs:
        kwargs.pop("transform")
    if "color" not in kwargs:
        kwargs["color"] = "grey"
    if "linewidths" not in kwargs:
        kwargs["linewidths"] = 0.5
    if "linestyles" not in kwargs:
        kwargs["linestyles"] = "dashed"

    ax.vlines(positions, 0, 1, transform=ax.get_xaxis_transform(), **kwargs)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Input-output for the files related with this package.
"""

__all__ = [
    "load_template",
    "read_template",
    "dump_spinham_txt",
    "dump_pickle",
    "load_pickle",
]

import numpy as np

from radtools.decorate.array import print_2d_array
from radtools.decorate.stats import logo
from radtools.spinham.constants import TXT_FLAGS
from radtools.spinham.hamiltonian import SpinHamiltonian
from radtools.spinham.template import ExchangeTemplate

meV_TO_J = 1.602176634e-22


def dump_pickle(object, filename):
    """
    Save any python Object in a binary format.

    Parameters
    ----------
    filename : str
        Name of the file for the Object to be saved in.
        ".pickle" is added automatically.
    """
    import pickle

    with open(f"{filename}.pickle", "wb") as file:
        pickle.dump(object, file)


def load_pickle(filename):
    r"""
    Load any python Object from a binary format.

    Parameters
    ----------
    filename : str
        Name of the .pickle file for the Object to be loaded from.
    """

    import pickle

    with open(filename, "rb") as file:
        object = pickle.load(file)
    return object


def load_template(filename):
    r"""
    Read template from the template file.

    .. versionchanged:: 0.8.1 Renamed from ``read_template``

    See :ref:`template-draft` for the description of the template file format.

    Parameters
    ----------
    filename : str
        Path to the template file.

    Returns
    -------
    template : :py:class:`.ExchangeTemplate`
        Exchange template.

    Notes
    -----
    See also :ref:`Template specification <api>`.
    """

    # Constants
    major_sep = "=" * 20
    minor_sep = "-" * 20
    neighbors_flag = "Neighbors template:"

    template = ExchangeTemplate()

    with open(filename) as file:
        line = True
        while line:
            line = file.readline()

            # Read the neighbors names
            if line and neighbors_flag in line:
                while line and major_sep not in line:
                    if line and minor_sep in line:
                        line = file.readline()
                        name = line.split()[0]
                        try:
                            latex_name = line.split()[1]
                        except IndexError:
                            latex_name = name
                        template.names[name] = []
                        template.latex_names[name] = latex_name
                        line = file.readline()
                        while line and minor_sep not in line and major_sep not in line:
                            atom1 = line.split()[0]
                            atom2 = line.split()[1]

                            R = tuple(map(int, line.split()[2:]))
                            template.names[name].append((atom1, atom2, R))
                            line = file.readline()
                    if line and major_sep in line:
                        break
                    if line and minor_sep not in line:
                        line = file.readline()
    return template


# For backward compatibility
read_template = load_template


def dump_spinham_txt(
    spinham: SpinHamiltonian,
    filename=None,
    anisotropic=True,
    matrix=True,
    dmi=True,
    template=None,
    decimals=4,
    additional_stats=None,
):
    """
    Save the :py:class:`.SpinHamiltonian` in a human-readable format.

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian to be saved.
    filename : str, optional
        Name of the file for the Hamiltonian to be saved in.
        If not given, the Hamiltonian will be printed in the console.
    anisotropic : bool, default True
        Whether to output anisotropic exchange.
    matrix : bool, default True
        Whether to output whole matrix exchange.
    dmi : bool, default True
        Whether to output DMI exchange.
    template : :py:class:`.ExchangeTemplate`, optional
        If provided, then not the SpinHamiltonian will be written, but the model
        based on the template.
    decimals : int, default 4
        Number of decimals to be printed (only for the exchange values).
    additional_stats : str, optional
        Additional info, which will be printed right after the logo, before the
        main separator.
    """

    main_separator = "=" * 80 + "\n"
    separator = "-" * 80 + "\n"
    spinham_txt = []
    fmt = f"{decimals+4}.{decimals}f"

    spinham_txt.append(main_separator)
    spinham_txt.append(logo(date_time=True, line_length=80) + "\n")
    if additional_stats is not None:
        spinham_txt.append(additional_stats)
    spinham_txt.append(main_separator)
    spinham_txt.append(TXT_FLAGS["cell"] + "\n")
    spinham_txt.append(
        print_2d_array(
            spinham.cell,
            borders=False,
            fmt="^.8f",
            print_result=False,
            header_row=["x", "y", "z"],
        )
        + "\n"
    )
    spinham_txt.append(main_separator)
    spinham_txt.append(TXT_FLAGS["atoms"] + "\n")
    header_column = []
    header_row = [
        f"Index Name",
        "a1 (rel)",
        "a2 (rel)",
        "a3 (rel)",
        "x",
        "y",
        "z",
    ]
    atom_data = np.zeros((len(spinham.atoms), 6))
    for a_i, atom in enumerate(spinham.atoms):
        header_column.append(f"{atom.index:<5} {atom.name:<4}")
        atom_data[a_i, :3] = spinham.get_atom_coordinates(atom, relative=True)
        atom_data[a_i, 3:] = spinham.get_atom_coordinates(atom, relative=False)
    spinham_txt.append(
        print_2d_array(
            atom_data,
            borders=False,
            fmt="^.8f",
            print_result=False,
            header_row=header_row,
            header_column=header_column,
        )
        + "\n"
    )
    spinham_txt.append(main_separator)

    write_spins = True
    write_magmoms = True
    for atom in spinham.magnetic_atoms:
        try:
            atom.magmom
        except ValueError:
            write_magmoms = False
        try:
            atom.spin
            atom.spin_vector
        except ValueError:
            write_spins = False
    if write_magmoms:
        spinham_txt.append(TXT_FLAGS["magmoms"] + "\n")
        header_column = []
        header_row = [f"Index Name", "m_x", "m_y", "m_z"]
        magmom_data = np.zeros((len(spinham.magnetic_atoms), 3))
        for a_i, atom in enumerate(spinham.magnetic_atoms):
            header_column.append(f"{atom.index:<5} {atom.name:<4}")
            magmom_data[a_i] = atom.magmom
        spinham_txt.append(
            print_2d_array(
                magmom_data,
                borders=False,
                fmt="^.8f",
                print_result=False,
                header_row=header_row,
                header_column=header_column,
            )
            + "\n"
        )
        spinham_txt.append(main_separator)
    if write_spins:
        spinham_txt.append(TXT_FLAGS["spins"] + "\n")
        header_column = []
        header_row = [f"Index Name", "S", "S_x", "S_y", "S_z"]
        spin_data = np.zeros((len(spinham.magnetic_atoms), 4))
        for a_i, atom in enumerate(spinham.magnetic_atoms):
            header_column.append(f"{atom.index:<5} {atom.name:<4}")
            spin_data[a_i, 0] = atom.spin
            spin_data[a_i, 1:] = atom.spin_vector
        spinham_txt.append(
            print_2d_array(
                spin_data,
                borders=False,
                fmt="^.8f",
                print_result=False,
                header_row=header_row,
                header_column=header_column,
            )
            + "\n"
        )
        spinham_txt.append(main_separator)

    spinham_txt.append(TXT_FLAGS["notation"] + "\n")
    spinham_txt.append(f"{spinham.notation}\n")
    spinham_txt.append(spinham.notation_string + "\n")
    spinham_txt.append(main_separator)

    if template is None:
        spinham_txt.append(TXT_FLAGS["spinham"] + "\n")
        spinham_txt.append(
            f"{'Atom1':6} {'Atom2':6} (  i,   j,   k) {'J_iso':^{decimals+4}} {'Distance':^8}\n"
        )
        bonds_data = []
        for atom1, atom2, (i, j, k), J in spinham:
            data_entry = []
            distance = spinham.get_distance(atom1, atom2, (i, j, k))
            atom1 = f"{atom1.name}({atom1.index})"
            atom2 = f"{atom2.name}({atom2.index})"
            data_entry.append(separator)
            data_entry.append(
                f"{atom1:6} {atom2:6} "
                + f"({i:>3}, {j:>3}, {k:>3}) "
                + f"{J.iso:^{decimals+4}.{decimals}f} "
                + f"{distance:^8.4f}\n"
            )
            if matrix:
                data_entry.append(TXT_FLAGS["matrix"] + "\n")
                data_entry.append(
                    print_2d_array(
                        J.matrix, fmt=fmt, borders=False, print_result=False, shift=2
                    )
                    + "\n"
                )
            if anisotropic:
                data_entry.append(TXT_FLAGS["aniso"] + "\n")
                data_entry.append(
                    print_2d_array(
                        J.aniso, fmt=fmt, borders=False, print_result=False, shift=2
                    )
                    + "\n"
                )
            if dmi:
                data_entry.append(TXT_FLAGS["dmi_module"] + "\n")
                data_entry.append(f"  {J.dmi_module:{decimals+4}.{decimals}f}\n")
                data_entry.append(TXT_FLAGS["dmi_relative"] + "\n")
                data_entry.append(f"  {J.rel_dmi:{decimals+4}.{decimals}f}\n")
                data_entry.append(TXT_FLAGS["dmi"] + "\n")
                data_entry.append(
                    print_2d_array(
                        J.dmi, fmt=fmt, borders=False, print_result=False, shift=2
                    )
                    + "\n"
                )
            bonds_data.append(["".join(data_entry), distance])
        bonds_data = sorted(bonds_data, key=lambda x: x[1])
        bonds_data = [x[0] for x in bonds_data]
        spinham_txt.append("".join(bonds_data))
    else:
        spinham = spinham.formed_model(template)
        spinham_txt.append(TXT_FLAGS["spinmodel"] + "\n")
        for parameter in template.names:
            bonds = template.names[parameter]
            J = spinham[bonds[0]]
            spinham_txt.append(separator)
            spinham_txt.append(f"{parameter} contains {len(bonds)} bonds\n")
            spinham_txt.append(TXT_FLAGS["iso"])
            spinham_txt.append(f"  {J.iso:{decimals+4}.{decimals}f}\n")
            if matrix:
                spinham_txt.append(TXT_FLAGS["matrix"] + "\n")
                spinham_txt.append(
                    print_2d_array(
                        J.matrix, fmt=fmt, borders=False, print_result=False, shift=2
                    )
                    + "\n"
                )
            if anisotropic:
                spinham_txt.append(TXT_FLAGS["aniso"] + "\n")
                spinham_txt.append(
                    print_2d_array(
                        J.aniso, fmt=fmt, borders=False, print_result=False, shift=2
                    )
                    + "\n"
                )
            if dmi:
                spinham_txt.append(TXT_FLAGS["dmi_module"] + "\n")
                spinham_txt.append(f"  {J.dmi_module:{decimals+4}.{decimals}f}\n")
                spinham_txt.append(TXT_FLAGS["dmi_relative"] + "\n")
                spinham_txt.append(f"  {J.rel_dmi:{decimals+4}.{decimals}f}\n")
                spinham_txt.append(TXT_FLAGS["dmis"] + "\n")
                for atom1, atom2, (i, j, k) in bonds:
                    J = spinham[atom1, atom2, (i, j, k)]
                    atom1 = spinham.get_atom(atom1)
                    atom2 = spinham.get_atom(atom2)
                    atom1 = f"{atom1.name}({atom1.index})"
                    atom2 = f"{atom2.name}({atom2.index})"
                    spinham_txt.append(
                        print_2d_array(
                            J.dmi, fmt=fmt, borders=False, print_result=False, shift=2
                        )
                    )
                    spinham_txt.append(
                        f"   ({atom1:6} {atom2:6} {i:>3}, {j:>3}, {k:>3})\n"
                    )
            else:
                spinham_txt.append(TXT_FLAGS["bonds"] + "\n")
                spinham_txt.append(f"   {'Atom1':6} {'Atom2':6}   i,   j,   k\n")
                for atom1, atom2, (i, j, k) in bonds:
                    atom1 = spinham.get_atom(atom1)
                    atom2 = spinham.get_atom(atom2)
                    atom1 = f"{atom1.name}({atom1.index})"
                    atom2 = f"{atom2.name}({atom2.index})"
                    spinham_txt.append(
                        f"   {atom1:6} {atom2:6} {i:>3}, {j:>3}, {k:>3}\n"
                    )
    spinham_txt.append(main_separator)
    spinham_txt = "".join(spinham_txt)
    if filename is not None:
        with open(filename, "w", encoding="utf-8") as file:
            file.write(spinham_txt)
    else:
        print(spinham_txt)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
PDOS
"""

import re
from copy import deepcopy

import numpy as np


class PDOS:
    r"""
    Partial density of states, projected on arbitrary projections.

    Supports k-resolved density of states.
    Support spin-polarised and spin-unpolarised cases.

    PDOS class is iterable (over :py:attr:`.projectors`) and
    supports item call (return PDOS for ``key`` projector).

    Operations of addition and subtraction are defined.

    Parameters
    ----------
    energy : |array_like|_
        Values of energy for the PDOS. Shape :math:`n_e` is assumed.
    pdos : |array_like|_
        Array with the values of PDOS. Shape is assumed to be:

        * Spin-polarized, k-resolved: :math:`(n, 2, n_k, n_e)`
        * Spin-unpolarized, k-resolved: :math:`(n, n_k, n_e)`
        * Spin-polarized, non k-resolved: :math:`(n, 2, n_e)`
        * Spin-unpolarized, non k-resolved: :math:`(n, n_e)`

        where :math:`n` is the number of projections,
        :math:`n_k` is the number of k-points,
        :math:`n_e` is the number of energy points.
    projectors_group : str
        Name of the projectors group.
    projectors : list
        Names of the projectors.
        If :py:attr:`projectors_group` has the form "l" or "l_j",
        where l is "s", "p", "d", "f" and j is the total angular momentum,
        the projectors are assigned automatically,
        otherwise it is necessary to provide :math:`n` projectors manually.
        The names of projectors are directly used in the plots.
    ldos : |array_like|_, optional
        Local density of states. Sum of partial density of states over all projectors.
        Computed based on ``pdos`` if not provided.
        Shape is assumed to be:

        * Spin-polarized, k-resolved: :math:`(2, n_k, n_e)`
        * Spin-unpolarized, k-resolved: :math:`(n_k, n_e)`
        * Spin-polarized, non k-resolved: :math:`(2, n_e)`
        * Spin-unpolarized, non k-resolved: :math:`(n_e)`

        where :math:`n_k` is the number of k-points,
        :math:`n_e` is the number of energy points.
    spin_pol : bool, default False
        Whether PDOS is spin-polarized or not.

    Attributes
    ----------
    energy : :numpy:`ndarray`
        Values of energy for the PDOS. Has the shape :math:`n_e`.
    ldos : :numpy:`ndarray`
    pdos : :numpy:`ndarray`
    projectors_group : str
        Name of the projectors group.
    projectors : list
        Names of the projectors.
    spin_pol : bool, default False
        Whether PDOS is spin-polarized or not.
    k_resolved : bool, default False
    """

    def __init__(
        self,
        energy,
        pdos,
        projectors_group: str,
        projectors,
        ldos=None,
        spin_pol=False,
    ):
        self.energy = np.array(energy)
        self._pdos = None
        self._ldos = None
        self.projectors_group = projectors_group
        self.projectors = projectors
        self.spin_pol = spin_pol

        self.pdos = pdos
        self.ldos = ldos

    def __add__(self, other):
        if not isinstance(other, PDOS):
            raise TypeError(
                f"Addition is not supported between "
                + f"{type(self)} and {type(other)}"
            )
        if (
            self._pdos.shape == other._pdos.shape
            and self.spin_pol == other.spin_pol
            and self.projectors_group == other.projectors_group
            and set(self.projectors) == set(other.projectors)
        ):
            pdos = self._pdos + other._pdos
            ldos = self._ldos + other._ldos

            return PDOS(
                energy=self.energy,
                pdos=pdos,
                projectors_group=self.projectors_group,
                projectors=self.projectors,
                ldos=ldos,
                spin_pol=self.spin_pol,
            )
        else:
            raise ValueError(
                "There is a mismatch between self and other:\n"
                + f"    pdos shape: {self._pdos.shape} {other._pdos.shape}\n"
                + f"    spin_pol: {self.spin_pol} {other.spin_pol}\n"
                + f"    projectors_group: {self.projectors_group} {other.projectors_group}\n"
                + f"    projectors: {self.projectors} {other.projectors}\n"
            )

    def __sub__(self, other):
        if not isinstance(other, PDOS):
            raise TypeError(
                f"Subtraction is not supported between "
                + f"{type(self)} and {type(other)}"
            )
        if (
            self._pdos.shape == other._pdos.shape
            and self.spin_pol == other.spin_pol
            and self.projectors_group == other.projectors_group
            and set(self.projectors) == set(other.projectors)
        ):
            pdos = self._pdos - other._pdos
            ldos = self._ldos - other._ldos
            return PDOS(
                energy=self.energy,
                pdos=pdos,
                projectors_group=self.projectors_group,
                projectors=self.projectors,
                ldos=ldos,
                spin_pol=self.spin_pol,
            )
        else:
            raise ValueError(
                "There is a mismatch between self and other:\n"
                + f"    pdos shape: {self._pdos.shape} {other._pdos.shape}\n"
                + f"    spin_pol: {self.spin_pol} {other.spin_pol}\n"
                + f"    projectors_group: {self.projectors_group} {other.projectors_group}\n"
                + f"    projectors: {self.projectors} {other.projectors}\n"
            )

    def __iter__(self):
        return PDOSIterator(self)

    def __len__(self):
        return len(self.projectors)

    def __contains__(self, item):
        return item in self.projectors

    def __getitem__(self, key) -> np.ndarray:
        r"""
        Return pdos by the projector name.

        Parameters
        ----------
        key : str or int
            Projector`s name or index.
        Returns
        -------
        pdos : :numpy:`ndarray`
            Partial density of states.
        """
        if isinstance(key, str):
            key = self.projectors.index(key)

        return self.pdos[key]

    @property
    def ldos(self) -> np.ndarray:
        r"""
        Local density of states.

        Returns
        -------
        ldos : :numpy:`ndarray`
            Summed density of states along all projectors.
            Has the following shapes:

            * Spin-polarized, k-resolved: :math:`(2, n_k, n_e)`
            * Spin-unpolarized, k-resolved: :math:`(n_k, n_e)`
            * Spin-polarized, non k-resolved: :math:`(2, n_e)`
            * Spin-unpolarized, non k-resolved: :math:`(n_e)`

            where :math:`n_k` is the number of k-points,
            :math:`n_e` is the number of energy points.
        """

        return self._ldos

    @ldos.setter
    def ldos(self, new_ldos):
        if (
            self._pdos is not None
            and new_ldos is not None
            and np.array(new_ldos).shape != self.pdos.shape[1:]
        ):
            raise ValueError(
                f"New LDOS shape {new_ldos.shape} does not match "
                + f"PDOS shape ({self.pdos.shape[1:]})"
            )

        if new_ldos is not None:
            self._ldos = np.array(new_ldos)
        else:
            self._ldos = np.sum(self._pdos, axis=0)

        if len(self.energy) != self.ldos.shape[-1]:
            raise ValueError(
                f"LDOS does not match with energy: "
                + f"{len(self.energy)} energy points, {self.ldos.shape[-1]} LDOS points"
            )

    @property
    def pdos(self) -> np.ndarray:
        r"""
        Partial density of states.

        Returns
        -------
        pdos : :numpy:`ndarray`
            Has the following shapes:

            * Spin-polarized, k-resolved: :math:`(n, 2, n_k, n_e)`
            * Spin-unpolarized, k-resolved: :math:`(n, n_k, n_e)`
            * Spin-polarized, non k-resolved: :math:`(n, 2, n_e)`
            * Spin-unpolarized, non k-resolved: :math:`(n, n_e)`

            where :math:`n` is the number of projections,
            :math:`n_k` is the number of k-points,
            :math:`n_e` is the number of energy points.
        """

        return self._pdos

    @pdos.setter
    def pdos(self, new_pdos):
        new_pdos = np.array(new_pdos)
        if self._ldos is not None and new_pdos.shape[1:] != self.ldos.shape:
            raise ValueError(
                f"New PDOS shape {new_pdos.shape[1:]} does not match "
                + f"LDOS shape ({self.ldos.shape})"
            )
        self._pdos = new_pdos

        if len(self.energy) != self.pdos.shape[-1]:
            raise ValueError(
                f"PDOS does not match with energy: "
                + f"{len(self.energy)} energy points, {self.pdos.shape[-1]} PDOS points"
            )

        if len(self.projectors) != self.pdos.shape[0]:
            raise ValueError(
                f"PDOS does not match with projectors: "
                + f"{len(self.projectors)} projectors, {self.pdos.shape[0]} PDOS"
            )

    @property
    def k_resolved(self):
        r"""
        Check if pdos is k-resolved based on shape of :py:attr:`.pdos`.
        """

        return (
            self.spin_pol
            and len(self.pdos.shape) == 4
            or not self.spin_pol
            and len(self.pdos.shape) == 3
        )

    @property
    def kpoints(self):
        r"""
        Return k-points.

        Returns
        -------
        kpoints : :numpy:`ndarray`
            K-points.
        """

        if self.k_resolved:
            return np.linspace(1, self.pdos.shape[-2], self.pdos.shape[-2])
        else:
            return np.array([1])

    @property
    def n_k(self):
        r"""
        Return number of k-points.

        Returns
        -------
        n_k : int
            Number of k-points.
        """

        if self.k_resolved:
            return self.pdos.shape[-2]
        else:
            return 1

    @property
    def n_e(self):
        r"""
        Return number of energy points.

        Returns
        -------
        n_e : int
            Number of energy points.
        """

        return self.pdos.shape[-1]

    def squeeze(self):
        r"""
        Squeeze k-resolved PDOS.

        Sum the pdos over kpoints and divide by the number of kpoints.

        See Also
        --------
        squeezed : Returns new object.

        Notes
        -----
        It modifies the instance on which called.
        """

        if self.k_resolved:
            self._pdos = np.sum(self._pdos, axis=1 + int(self.spin_pol)) / self.n_k
            self._ldos = np.sum(self._ldos, axis=int(self.spin_pol)) / self.n_k

    def squeezed(self):
        r"""
        Return new instance with squeezed PDOS.

        Calls :py:func:`.PDOS.squeeze`.

        Returns
        -------
        pdos_squeezed : :py:class:`.PDOS`
            Squeezed PDOS.

        See Also
        --------
        squeeze : Modifies current object.
        """

        squeezed_pdos = deepcopy(self)
        squeezed_pdos.squeeze()
        return squeezed_pdos

    def normalize(self, zeros_to_none=False):
        r"""
        Normalize values of PDOS to 1 for each k and energy point.

        If :math:`x_i(E, k)` is  PDOS of the projector :math:`i`,
        then after this function does the following:

        .. math::
            x_i(E, k) \rightarrow \dfrac{x_i(E, k)}{\sum_{j=0}^{n} x_j(E, k)}

        where :math:`n` is the total number of projectors.
        Those sums are computed individually for spin-up and
        spin-down in the spin-polarized case.

        Parameters
        ----------
        zeros_to_none : bool, default False
            If True, then the values of PDOS and LDOS > 1e-8
            will be replaced with ``None``.

        See Also
        --------
        normalized : Returns new object.

        Notes
        -----
        It modifies the instance on which called.

        """
        tmp_ldos = np.where(self._ldos > 1e-8, self._ldos, 1)
        for i in range(0, self.pdos.shape[0]):
            self._pdos[i] = self._pdos[i] / tmp_ldos
            if zeros_to_none:
                self._pdos[i] = np.where(self._pdos[i] > 1e-8, self._pdos[i], None)
        self._ldos = self._ldos / tmp_ldos

        if zeros_to_none:
            self._ldos = np.where(self._ldos > 1e-8, self._ldos, None)

    def normalized(self, zeros_to_none=False):
        r"""
        Return new instance with normalized PDOS.

        Calls :py:func:`PDOS.normalize`.

        Parameters
        ----------
        zeros_to_none : bool, default False
            If True, then the values of PDOS and LDOS equal to zero
            will be replaced with ``None``.

        Returns
        -------
        normalized_pdos : :py:class:`PDOS`
            Normalized PDOS.

        See Also
        --------
        normalize : Modifies current object.
        """

        normalized_pdos = deepcopy(self)
        normalized_pdos.normalize(zeros_to_none=zeros_to_none)
        return normalized_pdos

    def dump_txt(self, output_name):
        r"""
        Save PDOS as .txt file.

        First line is a header.
        """

        header = ""
        fmt = ""
        if self.k_resolved:
            header += "# ik  "
            fmt += "%5.0f "
        header += "E (eV)     "
        fmt += "%10.3f "

        if self.spin_pol:
            header += "ldosup(E) ldosdw(E) "
            fmt += "%9.3E %9.3E "
            for i in self.projectors:
                n = max(9, len(i) + 2)
                header += f"{i+'up':>{n}} {i+'dw':>{n}} "
                fmt += f"%{n}.3E %{n}.3E "
        else:
            header += "ldos(E)   "
            fmt += "%9.3E "
            for i in self.projectors:
                n = max(9, len(i))
                header += f"{i:>{n}} "
                fmt += f"%{n}.3E "

        data = []
        nepoints = self.pdos.shape[-1]
        if self.k_resolved:
            if self.spin_pol:
                nkpoints = self.pdos.shape[2]
                ldos = self.ldos.reshape(2, nkpoints * nepoints)
                pdos = self.pdos.reshape(len(self.projectors), 2, nkpoints * nepoints)
            else:
                nkpoints = self.pdos.shape[1]
                ldos = self.ldos.reshape(nkpoints * nepoints)
                pdos = self.pdos.reshape(len(self.projectors), nkpoints * nepoints)
            data.append(np.repeat(np.linspace(1, nkpoints, nkpoints), nepoints))
            data.append(np.tile(self.energy, nkpoints))
        else:
            data.append(self.energy)
            ldos = self.ldos
            pdos = self.pdos

        if self.spin_pol:
            data.append(ldos[0])
            data.append(ldos[1])
            for i in range(pdos.shape[0]):
                data.append(pdos[i][0])
                data.append(pdos[i][1])
        else:
            data.append(ldos)
            for i in range(pdos.shape[0]):
                data.append(pdos[i])
        np.savetxt(output_name, np.array(data).T, fmt=fmt, header=header, comments="")


class PDOSIterator:
    def __init__(self, pdos: PDOS) -> None:
        self._list = pdos.projectors
        self._index = 0

    def __next__(self) -> str:
        if self._index < len(self._list):
            result = self._list[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self):
        return self


class PDOSQE(PDOS):
    r"""
    PDOS wrapper for |QE|_ pdos.

    Supports the order of projectors of |projwfc|_ (s,p,d,f) and
    the case of projection in the spin-orbit calculations.
    In the custom cases it is necessary to specify projectors manually.
    If :py:attr:`.projectors_group` has the form "l" or "l_j",
    where l is "s", "p", "d", "f" and j is the total angular momentum,
    the projectors are assigned automatically,
    otherwise it is necessary to provide :math:`n` projectors manually.
    The names of projectors are directly used in the plots.
    If :py:attr:`.projectors_group` is one of "s", "p", "d", "f", then the projectors are:

    * s : :math:`s`
    * p : :math:`p_z`, :math:`p_y`, :math:`p_x`
    * d : :math:`d_{z^2}`, :math:`d_{zx}`, :math:`d_{zy}`, :math:`d_{x^2 - y^2}`, :math:`d_{xy}`
    * f : :math:`f_{z^3}`, :math:`f_{yz^2}`, :math:`f_{xz^2}`, :math:`f_{z(x^2 - y^2)}`, :math:`f_{xyz}`, :math:`f_{y(3x^2 - y^2)}`, :math:`f_{x(x^2 - 3y^2)}`

    If :py:attr:`.projectors_group` has the form "l_j", then the projectors are :math:`(1, ..., 2j+1)`
    """

    _pattern = "[spdf]_j[0-9.]*"
    _projectors = {
        "s": ["s"],
        "p": ["$p_z$", "$p_y$", "$p_x$"],
        "d": ["$d_{z^2}$", "$d_{zx}$", "$d_{zy}$", "$d_{x^2 - y^2}$", "$d_{xy}$"],
        "f": [
            "$f_{z^3}$",
            "$f_{yz^2}$",
            "$f_{xz^2}$",
            "$f_{z(x^2 - y^2)}$",
            "$f_{xyz}$",
            "$f_{y(3x^2 - y^2)}$",
            "$f_{x(x^2 - 3y^2)}$",
        ],
    }

    def __init__(
        self,
        energy,
        pdos,
        projectors_group: str,
        projectors=None,
        ldos=None,
        spin_pol=False,
    ):
        if projectors is not None:
            pass
        elif projectors_group in self._projectors:
            projectors = self._projectors[projectors_group]
        elif re.fullmatch(self._pattern, projectors_group):
            l, j = projectors_group.split("_j")
            m_j = range(0, int(2 * float(j) + 1))
            projectors = [f"{l} ($m_J = {i - float(j):>4.1f}$)" for i in m_j]
            projectors_group = f"{l} (J = {j})"
        else:
            raise ValueError(
                "Projectors can not be assigned automatically, "
                + "you have to provide explicit list of projectors. "
                + f"Projectors group: {projectors_group}"
            )
        super().__init__(energy, pdos, projectors_group, projectors, ldos, spin_pol)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Exchange parameter class.
"""
__all__ = ["ExchangeParameter"]
# SELF is introduced in python 3.11, it is too fresh for my taste
from typing import TypeVar

Self = TypeVar("Self", bound="ExchangeParameter")

import numpy as np

from radtools.decorate.array import print_2d_array


class ExchangeParameter:
    r"""
    Exchange parameter (:math:`\boldsymbol{J}`) class.

    Basically it is a wrapper around :numpy:`ndarray`
    with predefined exchange-specific attributes.
    Any function, which work on :numpy:`ndarray` should work on
    :py:class:`.ExchangeParameter` as well. It will act on the
    :py:attr:`.matrix` attribute.

    If ``matrix`` is specified then ``iso``, ``aniso`` and ``dmi`` are
    ignored and derived from ``matrix``. If ``matrix`` is not specified
    then it is derived from ``iso``, ``aniso`` and ``dmi``.
    If nothing is provided exchange matrix is set to zero matrix.

    Parameters
    ----------
    iso : int or float, optional
        Value of isotropic exchange parameter.
    aniso : (3, 3) |array_like|_, optional
        3 x 3 matrix of symmetric anisotropic exchange.
    dmi : (3,) |array_like|_, optional
        Dzyaroshinsky-Moria interaction vector :math:`(D_x, D_y, D_z)`.
    matrix : (3, 3) |array_like|_, optional
        Exchange matrix.
    """

    def __init__(self, matrix=None, iso=None, aniso=None, dmi=None) -> None:
        self._matrix = np.zeros((3, 3), dtype=float)
        self._iso = 0.0
        self._aniso = np.zeros((3, 3), dtype=float)
        self._dmi = np.zeros(3, dtype=float)

        if matrix is None:
            if iso is not None:
                self.iso = iso
            if dmi is not None:
                self.dmi = dmi
            if aniso is not None:
                self.aniso = aniso
        else:
            self.matrix = matrix

    def __format__(self, fmt):
        return self.__str__(fmt=fmt)

    def __str__(self, fmt=".4f"):
        return print_2d_array(self.matrix, fmt=fmt, print_result=False, borders=False)

    def __repr__(self):
        return f"ExchangeParameter({self.matrix.__repr__()})"

    @property
    def __array_interface__(self):
        return self.matrix.__array_interface__

    @property
    def T(self):
        r"""
        Transposes a matrix of the exchange parameter.

        It returns new instance of the :py:class`.ExchangeParameter`
        with transposed matrix.

        Returns
        -------
        ExchangeParameter : :py:class`.ExchangeParameter`
            New instance of the :py:class`.ExchangeParameter`
            with transposed matrix.
        """
        return ExchangeParameter(matrix=self.matrix.T)

    @property
    def matrix(self) -> np.ndarray:
        r"""
        Full exchange matrix.

        .. code-block:: python

            [[J_xx, J_xy, J_xz],
             [J_yx, J_yy, J_yz],
             [J_zx, J_zy, J_zz]]

        Returns
        -------
        matrix : (3, 3) :numpy:`ndarray`
            Full exchange matrix.

        See Also
        --------
        symm_matrix
            Symmetric part.
        asymm_matrix
            Asymmetric part.
        """

        return self._matrix

    @matrix.setter
    def matrix(self, new_matrix):
        try:
            new_matrix = np.array(new_matrix, dtype=float)
        except:
            raise ValueError(f"New matrix is not array_like: \n{new_matrix}")
        if new_matrix.shape != (3, 3):
            raise ValueError("Matrix shape has to be equal to (3, 3)")
        self._matrix = new_matrix

    @property
    def symm_matrix(self) -> np.ndarray:
        r"""
        Symmetric part of exchange :py:attr:`.matrix`.

        .. math::

            J_{symm} = \dfrac{J + J^T}{2}

        Returns
        -------
        symm_matrix : (3, 3) :numpy:`ndarray`
            Symmetric part of exchange matrix.

        See Also
        --------
        asymm_matrix
            Asymmetric part.
        matrix
            Full matrix.
        """

        return (self.matrix + self.matrix.T) / 2

    @property
    def asymm_matrix(self) -> np.ndarray:
        """
        Asymmetric part of exchange :py:attr:`.matrix`.

        .. math::

            J_{asymm} = \dfrac{J - J^T}{2}

        Returns
        -------
        asymm_matrix : (3, 3) :numpy:`ndarray`
            Asymmetric part of exchange matrix.

        See Also
        --------
        symm_matrix
            Symmetric part.
        matrix
            Full matrix.
        """

        return (self.matrix - self.matrix.T) / 2

    @property
    def iso(self) -> float:
        r"""
        Value of isotropic exchange parameter.

        Derived from the exchange matrix (:math:`\mathbf{J}`) as

        .. math::
            J_{iso} = \dfrac{1}{3}Tr(\mathbf{J})

        Returns
        -------
        iso : float
            Value of isotropic exchange parameter.

        See Also
        --------
        iso_matrix
            Isotropic exchange in a matrix form.
        """

        return np.trace(self.symm_matrix) / 3

    @iso.setter
    def iso(self, new_iso):
        new_iso = new_iso - self.iso
        self.matrix = self.matrix + (new_iso) * np.identity(3, dtype=float)

    @property
    def iso_matrix(self) -> np.ndarray:
        r"""
        Isotropic part of the exchange :py:attr:`.matrix`.

        Matrix form:

        .. code-block:: python

            [[J, 0, 0],
             [0, J, 0],
             [0, 0, J]]

        Returns
        -------
        iso_matrix : (3, 3) :numpy:`ndarray`
            Isotropic part of the exchange matrix.

        See Also
        --------
        iso
            Isotropic exchange parameter.
        """

        return self.iso * np.identity(3, dtype=float)

    @property
    def aniso(self) -> np.ndarray:
        r"""
        3 x 3 matrix of symmetric anisotropic exchange.

        .. code-block:: python

            [[J_xx, J_xy, J_xz],
             [J_xy, J_yy, J_yz],
             [J_xz, J_yz, J_zz]]

        Derived from the exchange matrix (:math:`\mathbf{J}`) as

        .. math::
            \mathbf{J}_{aniso} = \mathbf{J}_{symm} - \dfrac{1}{3}Tr(\mathbf{J})
            \cdot \mathbf{I}

        where :math:`\mathbf{I}` is a :math:`3\times3` identity matrix.

        Returns
        -------
        aniso : (3, 3) :numpy:`ndarray`
            Matrix of symmetric anisotropic exchange.

        See Also
        --------
        aniso_diagonal
            Diagonal part of the symmetric anisotropic exchange.
        aniso_diagonal_matrix
            Diagonal part of the symmetric anisotropic exchange in a matrix form.
        """

        return self.symm_matrix - self.iso * np.identity(3, dtype=float)

    @aniso.setter
    def aniso(self, new_aniso):
        try:
            new_aniso = np.array(new_aniso, dtype=float)
        except:
            raise ValueError(f"New aniso is not array_like: \n{new_aniso}")
        if new_aniso.shape != (3, 3):
            raise ValueError("Aniso matrix shape have " + "to be equal to (3, 3)")
        # correction is correct (see matrix update)
        new_aniso = new_aniso - self.aniso
        self.matrix = self.matrix + new_aniso

    @property
    def aniso_diagonal(self) -> np.ndarray:
        r"""
        Diagonal part of the symmetric anisotropic exchange.

        .. code-block:: python

            [J_xx, J_yy, J_zz]

        Returns
        -------
        aniso_diagonal : (3,) :numpy:`ndarray`
            Diagonal part of the symmetric anisotropic exchange.

        See Also
        --------
        aniso
            3 x 3 matrix of symmetric anisotropic exchange.
        aniso_diagonal_matrix
            Diagonal part of the symmetric anisotropic exchange in a matrix form.
        """

        return np.diag(self.aniso)

    @property
    def aniso_diagonal_matrix(self) -> np.ndarray:
        r"""
        Diagonal part of the symmetric anisotropic exchange.

        .. code-block:: python

            [[J_xx, 0, 0],
             [0, J_yy, 0],
             [0, 0, J_zz]]

        Returns
        -------
        aniso_diagonal_matrix : (3, 3) :numpy:`ndarray`
            Diagonal part of the symmetric anisotropic exchange in matrix form.

        See Also
        --------
        aniso
            3 x 3 matrix of symmetric anisotropic exchange.
        aniso_diagonal
            Diagonal part of the symmetric anisotropic exchange.
        """

        return np.diag(np.diag(self.aniso))

    @property
    def dmi(self) -> np.ndarray:
        r"""
        Dzyaroshinsky-Moria interaction vector (Dx, Dy, Dz).

        .. code-block:: python

            [D_x, D_y, D_z]

        Returns
        -------
        dmi : (3,) :numpy:`ndarray`
            Dzyaroshinsky-Moria interaction vector.

        See Also
        --------
        dmi_matrix
            Dzyaroshinsky-Moria interaction in a matrix form.
        dmi_module
            Module of Dzyaroshinsky-Moria vector.
        rel_dmi
            Relative strength of Dzyaroshinsky-Moria interaction.
        """

        return np.array(
            [self.asymm_matrix[1][2], self.asymm_matrix[2][0], self.asymm_matrix[0][1]],
            dtype=float,
        )

    @dmi.setter
    def dmi(self, new_dmi):
        try:
            new_dmi = np.array(new_dmi, dtype=float)
        except:
            raise ValueError(f"New DMI is not array_like: \n{new_dmi}")
        if new_dmi.shape != (3,):
            raise ValueError(f"DMI have to be a 3 component vector. {new_dmi}")
        new_dmi = new_dmi - self.dmi
        dmi_matrix = np.array(
            [
                [0, new_dmi[2], -new_dmi[1]],
                [-new_dmi[2], 0, new_dmi[0]],
                [new_dmi[1], -new_dmi[0], 0],
            ],
            dtype=float,
        )
        self.matrix = self.matrix + dmi_matrix

    @property
    def dmi_matrix(self) -> np.ndarray:
        r"""
        Asymmetric part of the exchange :py:attr:`.matrix`.

        .. code-block:: python

            [[0, D_z, -D_y],
             [-D_z, 0, D_x],
             [D_y, -D_x, 0]]

        Returns
        -------
        dmi_matrix : (3, 3) :numpy:`ndarray`
            Asymmetric part of the exchange matrix.

        See Also
        --------
        dmi
            Dzyaroshinsky-Moria interaction vector.
        dmi_module
            Module of Dzyaroshinsky-Moria vector.
        rel_dmi
            Relative strength of Dzyaroshinsky-Moria interaction.
        """

        return self.asymm_matrix

    @property
    def dmi_module(self) -> float:
        r"""
        Length of the DMI vector in the units of exchange interaction.

        Returns
        -------
        dmi_module : float
            Length of the DMI vector in the units of exchange interaction.

        See Also
        --------
        dmi
            Dzyaroshinsky-Moria interaction vector.
        dmi_matrix
            Dzyaroshinsky-Moria interaction in a matrix form.
        rel_dmi
            Relative strength of Dzyaroshinsky-Moria interaction.
        """

        return np.linalg.norm(self.dmi)

    @property
    def rel_dmi(self) -> float:
        r"""
        Relative value of DMI.

        .. math::
            \frac{\vert DMI\vert}{\vert J\vert}

        Returns
        -------
        rel_dmi : float
            Relative value of DMI.

        See Also
        --------
        dmi
            Dzyaroshinsky-Moria interaction vector.
        dmi_matrix
            Dzyaroshinsky-Moria interaction in a matrix form.
        dmi_module
            Module of Dzyaroshinsky-Moria vector.
        """

        return self.dmi_module / abs(self.iso)

    @property
    def xx(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{xx}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.xx
            1.0
            >>> J.matrix[0][0]
            1.0

        Returns
        -------
        xx : float
            Value of exchange parameter :math:`J_{xx}`.
        """
        return self.matrix[0][0]

    @xx.setter
    def xx(self, new_xx):
        self._matrix[0][0] = float(new_xx)

    @property
    def xy(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{xy}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.xy
            2.0
            >>> J.matrix[0][1]
            2.0

        Returns
        -------
        xy : float
            Value of exchange parameter :math:`J_{xy}`.
        """
        return self.matrix[0][1]

    @xy.setter
    def xy(self, new_xy):
        self._matrix[0][1] = float(new_xy)

    @property
    def xz(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{xz}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.xz
            3.0
            >>> J.matrix[0][2]
            3.0

        Returns
        -------
        xz : float
            Value of exchange parameter :math:`J_{xz}`.
        """
        return self.matrix[0][2]

    @xz.setter
    def xz(self, new_xz):
        self._matrix[0][2] = float(new_xz)

    @property
    def yx(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{yx}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.yx
            4.0
            >>> J.matrix[1][0]
            4.0

        Returns
        -------
        yx : float
            Value of exchange parameter :math:`J_{yx}`.
        """
        return self.matrix[1][0]

    @yx.setter
    def yx(self, new_yx):
        self._matrix[1][0] = float(new_yx)

    @property
    def yy(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{yy}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.yy
            5.0
            >>> J.matrix[1][1]
            5.0

        Returns
        -------
        yy : float
            Value of exchange parameter :math:`J_{yy}`.
        """
        return self.matrix[1][1]

    @yy.setter
    def yy(self, new_yy):
        self._matrix[1][1] = float(new_yy)

    @property
    def yz(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{yz}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.yz
            6.0
            >>> J.matrix[1][2]
            6.0

        Returns
        -------
        yz : float
            Value of exchange parameter :math:`J_{yz}`.
        """
        return self.matrix[1][2]

    @yz.setter
    def yz(self, new_yz):
        self._matrix[1][2] = float(new_yz)

    @property
    def zx(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{zx}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.zx
            7.0
            >>> J.matrix[2][0]
            7.0

        Returns
        -------
        zx : float
            Value of exchange parameter :math:`J_{zx}`.
        """
        return self.matrix[2][0]

    @zx.setter
    def zx(self, new_zx):
        self._matrix[2][0] = float(new_zx)

    @property
    def zy(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{zy}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.zy
            8.0
            >>> J.matrix[2][1]
            8.0

        Returns
        -------
        zy : float
            Value of exchange parameter :math:`J_{zy}`.
        """
        return self.matrix[2][1]

    @zy.setter
    def zy(self, new_zy):
        self._matrix[2][1] = float(new_zy)

    @property
    def zz(self) -> float:
        r"""
        Value of exchange parameter :math:`J_{zz}`.

        .. versionadded:: 0.8.1

        .. doctest::

            >>> from radtools.spinham import ExchangeParameter
            >>> J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
            >>> J.zz
            9.0
            >>> J.matrix[2][2]
            9.0

        Returns
        -------
        zz : float
            Value of exchange parameter :math:`J_{zz}`.
        """
        return self.matrix[2][2]

    @zz.setter
    def zz(self, new_zz):
        self._matrix[2][2] = float(new_zz)

    # Definition of arithmetic operations5t

    # + (add)
    def __add__(self, other) -> Self:
        if isinstance(other, ExchangeParameter):
            matrix = self.matrix + other.matrix
        else:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for +: 'ExchangeParameter' and '{other.__class__.__name__}'"
            )
        return ExchangeParameter(matrix=matrix)

    # - (sub)
    def __sub__(self, other) -> Self:
        if isinstance(other, ExchangeParameter):
            matrix = self.matrix - other.matrix
        else:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for -: 'ExchangeParameter' and '{other.__class__.__name__}'"
            )
        return ExchangeParameter(matrix=matrix)

    # *
    def __mul__(self, number) -> Self:
        try:
            matrix = self.matrix * number
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for *: 'ExchangeParameter' and '{number.__class__.__name__}'"
            )
        return ExchangeParameter(matrix=matrix)

    # @
    def __matmul__(self, other) -> np.ndarray:
        try:
            result = self.matrix @ other
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for @: 'ExchangeParameter' and '{other.__class__.__name__}'"
            )
        return result

    # /
    def __truediv__(self, number) -> Self:
        try:
            matrix = self.matrix / number
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for /: 'ExchangeParameter' and '{number.__class__.__name__}'"
            )
        return ExchangeParameter(matrix=matrix)

    # //
    def __floordiv__(self, number) -> Self:
        try:
            matrix = self.matrix // number
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for //: 'ExchangeParameter' and '{number.__class__.__name__}'"
            )
        return ExchangeParameter(matrix=matrix)

    # %
    def __mod__(self, number) -> Self:
        try:
            matrix = self.matrix % number
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for %: 'ExchangeParameter' and '{number.__class__.__name__}'"
            )
        return ExchangeParameter(matrix=matrix)

    # *
    def __rmul__(self, number) -> Self:
        try:
            matrix = number * self.matrix
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for *: '{number.__class__.__name__}' and 'ExchangeParameter'"
            )
        return ExchangeParameter(matrix=matrix)

    # @
    def __rmatmul__(self, other) -> np.ndarray:
        try:
            result = other @ self.matrix
        except:
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for @: '{other.__class__.__name__}' and 'ExchangeParameter'"
            )
        return result

    # - (neg)
    def __neg__(self) -> Self:
        return ExchangeParameter(matrix=-self.matrix)

    # + (pos)
    def __pos__(self) -> Self:
        return ExchangeParameter(matrix=self.matrix)

    # abs()
    def __abs__(self) -> Self:
        return ExchangeParameter(matrix=np.abs(self.matrix))

    # ==
    def __eq__(self, other):
        if isinstance(other, ExchangeParameter):
            return (self.matrix == other.matrix).all()
        try:
            other = np.array(other, dtype=float)
        except:
            raise ValueError(f"Other is not array_like: \n{other}")
        if other.shape != (3, 3):
            raise ValueError("Other has to be equal to (3, 3)")
        return (self.matrix == other).all()

    # !=
    def __neq__(self, other):
        return not self == other


if __name__ == "__main__":
    J = ExchangeParameter(matrix=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    print(J.xy)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Exchange Module describes the spin Hamiltonian, defined on some :py:class:`.Crystal`.
"""


from .constants import *
from .hamiltonian import *
from .parameter import *
from .template import *

__all__ = []
__all__.extend(constants.__all__)
__all__.extend(hamiltonian.__all__)
__all__.extend(parameter.__all__)
__all__.extend(template.__all__)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
DOS
"""
import re
from os import walk
from os.path import join
from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
from termcolor import cprint

from radtools.decorate.axes import plot_hlines, plot_vlines
from radtools.dos.pdos import PDOSQE

PATTERN = "\\.pdos_atm\\#[0-9]*\\([a-zA-Z0-9]*\\)_wfc\\#[0-9]*\\([spdf_0-9j\\.]*\\)"


class DOSQE:
    r"""
    Analyse the folder with the |QE|_ results and store all information.

    Parameters
    ----------
    seedname : str
        Seedname for the |QE|_ output files.
    input_folder : str
        Path to the folder with |QE|_ output files.
    energy_window : tuple, optional
        Energy limits, necessary for correct plotting.

    Attributes
    ----------
    seedname : str
        Seedname for the |QE|_ output files.
    energy : array
        Energy values of the DOS/PDOS.
    k_resolved : bool
    case : int
    casename : str
        Human readable name of the case base on :py:attr:`.case`.
    nkpoints : int
        Number of k points for k resolved DOS. 1 if DOS is not k resolved.
    nepoints : int
        Number of energy points.
    filenames : list
    atoms : list
    """

    def __init__(
        self, seedname: str, input_folder: str, energy_window=None, efermi=0
    ) -> None:
        self.seedname = seedname
        self._input_folder = input_folder

        # Case and whether DOS is k resolved
        self._k_resolved = None
        self._case = None

        # Detect number of energy and k points and read energy values
        self.nkpoints = None
        self.nepoints = None
        self.energy = None
        self.efermi = efermi
        self._extract_energy()

        self._filenames = None

        self._atoms = {}
        self._wfcs = {}
        self._decompose_filenames()

        if energy_window is None:
            self.energy_window = (0, -1)
        else:
            i, j = 0, 0
            for i_e, e in enumerate(self.energy):
                if e <= energy_window[0]:
                    i += 1
                if e <= energy_window[1]:
                    j += 1
            self.energy_window = (i, min(j + 1, i_e))
        self.energy = self.energy[self.energy_window[0] : self.energy_window[1]]

    @property
    def spin_pol(self):
        return self.case in [2, 3]

    @property
    def case(self):
        r"""
        Detects case of the DOS calculations.

        Detects the case when called for the first time.

        There are 4 cases:

        #. Collinear
        #. Collinear, spin-polarized
        #. Non-collinear, non-spin-orbit
        #. Non-collinear, spin-orbit

        Headers of "seedname".pdos_tot according to the
        |projwfc|_:

        * Collinear:
            E DOS(E) PDOS(E)

        * Collinear, spin-polarized:
            E DOSup(E) DOSdw(E) PDOSup(E) PDOSdw(E)

        * Non-collinear, non-spin-orbit:
            E DOS(E) PDOSup(E) PDOSdw(E)

        * Non-collinear, spin-orbit:
            E DOS(E) PDOS(E)
        """

        # Detect the case
        if self._case is None:
            # Read header
            with open(
                join(f"{self._input_folder}", f"{self.seedname}.pdos_tot")
            ) as file:
                header = file.readline().lower().split()

            if "dos(e)" in header and "pdos(e)" in header:
                self._case = 1
            if "dos(e)" in header and "pdosup(e)" in header and "pdosdw(e)" in header:
                self._case = 3
            if (
                "dosup(e)" in header
                and "dosdw(e)" in header
                and "pdosup(e)" in header
                and "pdosdw(e)" in header
            ):
                self._case = 2
            # Differentiate between case 1 and 4
            if self._case == 1:
                # pattern is intentionally different from PATTERN
                pattern = (
                    "\\.pdos_atm\\#[0-9]*\\([a-zA-Z]*\\)_wfc\\#[0-9]*\\([spdf\\.]*\\)"
                )
                filename = self.filenames[0]
                if not re.fullmatch(
                    f"{re.escape(self.seedname)}{pattern}",
                    filename,
                ):
                    self._case = 4

            if self._case is None:
                raise RuntimeError(
                    "Unable to detect case, analysed header:\n"
                    + f"{header}\n"
                    + f"of the file:\n"
                    + f"{join(f'{self._input_folder}', f'{self.seedname}.pdos_tot')}"
                )

        return self._case

    @property
    def casename(self):
        r"""
        Human-readable name of the case.

        Returns
        -------
        case_name : str
            Name of the case.
        """
        case_names = [
            "collinear, spin-unpolarized",
            "collinear, spin-polarized",
            "non-collinear, non spin-orbit",
            "non-collinear, spin-orbit",
        ]
        return case_names[self.case - 1]

    @property
    def k_resolved(self):
        r"""
        Whether DOS is k-resolved.
        """

        if self._k_resolved is None:
            # Check for k-resolved
            with open(
                join(f"{self._input_folder}", f"{self.seedname}.pdos_tot")
            ) as file:
                self._k_resolved = "ik" in file.readline().lower().split()

        return self._k_resolved

    def _extract_energy(self):
        # Load data
        dos = np.loadtxt(
            join(f"{self._input_folder}", f"{self.seedname}.pdos_tot"), skiprows=1
        ).T

        # Load energy
        if self.k_resolved:
            # ik E ...
            self.nkpoints = int(dos[0][-1])
            self.nepoints = len(dos[0]) // self.nkpoints
            self.energy = dos[1][0 : self.nepoints] - self.efermi
        else:
            # E ...
            self.nkpoints = 1
            self.nepoints = len(dos[0])
            self.energy = dos[0] - self.efermi

    @property
    def filenames(self):
        r"""
        List of filenames (without "seedname".pdos_tot).
        """
        if self._filenames is None:
            # Get list of files in the folder
            all_filenames = []
            for _, _, filenames in walk(self._input_folder):
                all_filenames.extend(filenames)
                break

            self._filenames = []
            # Search
            for filename in all_filenames:
                if re.fullmatch(
                    f"{re.escape(self.seedname)}{PATTERN}",
                    filename,
                ):
                    self._filenames.append(filename)

        return self._filenames

    def _decompose_filenames(self):
        r"""
        Decompose filenames and extract information about atoms and projectors.

        Notes
        -----
        atoms : dist
            Dictionary with the atom labels and their numbers.

            .. code-block:: python

                _atoms = {atom_symbol: [atom_number, ...], ...}
        wfcs : dict
            Dictionary of projectors wave functions and their numbers.

            .. code-block:: python

                _wfcs = {(atom_symbol, atom_number): {wfc_symbol: [wfc_number, ...], ...}, ...}
        """

        for filename in self.filenames:
            # Detect names and numbers
            atom_number = int(filename.split(".pdos_atm#")[1].split("(")[0])
            atom_symbol = filename.split("(")[1].split(")")[0]
            wfc_number = int(filename.split(")_wfc#")[1].split("(")[0])
            wfc_symbol = filename.split(")_wfc#")[1].split("(")[1].split(")")[0]

            if atom_symbol not in self._atoms:
                self._atoms[atom_symbol] = []
            self._atoms[atom_symbol].append(atom_number)

            if (atom_symbol, atom_number) not in self._wfcs:
                self._wfcs[(atom_symbol, atom_number)] = {}
            if wfc_symbol not in self._wfcs[(atom_symbol, atom_number)]:
                self._wfcs[(atom_symbol, atom_number)][wfc_symbol] = []
            self._wfcs[(atom_symbol, atom_number)][wfc_symbol].append(wfc_number)

        # Sort entries
        for atom in self._atoms:
            self._atoms[atom] = list(set(self._atoms[atom]))
            self._atoms[atom].sort()
        for key in self._wfcs:
            for wfc_symbol in self._wfcs[key]:
                self._wfcs[key][wfc_symbol].sort()

    def __iter__(self):
        return DOSIterator(self)

    def __contains__(self, item):
        atom, atom_number, wfc, wfc_number = item
        return (
            atom in self.atoms
            and atom_number in self.atom_numbers(atom)
            and wfc in self.wfcs(atom, atom_number)
            and wfc_number in self.wfc_numbers(atom, wfc, atom_number)
        )

    def total_dos(self, squeeze=False):
        r"""
        Total density of states.

        Density of states, computed directly from plane-wave basis set.

        If DOS is k resolved, then sum over k points.

        Parameters
        ----------
        squeeze : bool, default False
            Whether to sum over k points. Ignored if DOS is not k resolved.
        Whether to fix updown in the noncollinear, non spin-orbit case.

        Returns
        -------
        total_dos : :numpy:`ndarray`
            Shape :math:`(2, n_e)` if DOS is collinear, spin-polarized;
            :math:`(n_e)` otherwise.
        """

        # Load data
        dos = np.loadtxt(
            join(self._input_folder, f"{self.seedname}.pdos_tot"), skiprows=1
        ).T

        if self.case == 2:
            # ik E DOS(up) DOS(down) PDOS(up) PDOS(down)
            if self.k_resolved:
                reshaped_dos = dos[2:4].reshape((2, self.nkpoints, self.nepoints))[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
                # Squeeze if necessary
                if squeeze:
                    return np.sum(reshaped_dos, axis=1) / self.nkpoints
                return reshaped_dos

            # E DOS(up) DOS(down) PDOS(up) PDOS(down)
            return dos[1:3][:, self.energy_window[0] : self.energy_window[1]]

        # ik E DOS PDOS
        # or
        # ik E DOS PDOS(up) PDOS(down)
        if self.k_resolved:
            reshaped_dos = dos[2].reshape((self.nkpoints, self.nepoints))[
                :, self.energy_window[0] : self.energy_window[1]
            ]
            # Squeeze if necessary
            if squeeze:
                return np.sum(reshaped_dos, axis=0) / self.nkpoints
            return reshaped_dos

        # E DOS PDOS
        # or
        # E DOS PDOS(up) PDOS(down)
        return dos[1][self.energy_window[0] : self.energy_window[1]]

    def total_pdos(self, squeeze=False):
        r"""
        Total partial density of states.

        If DOS is k resolved, then sum over k points.

        Parameters
        ----------
        squeeze : bool, default False
            Whether to sum over k points. Ignored if DOS is not k resolved.

        Returns
        -------
        total_pdos : :numpy:`ndarray`
            Shape :math:`(2, n_e)` if PDOS is collinear, spin-polarized;
            :math:`(n_e)` otherwise.
        """

        # Load data
        dos = np.loadtxt(
            join(f"{self._input_folder}", f"{self.seedname}.pdos_tot"), skiprows=1
        ).T

        if self.case == 2:
            # ik E DOS(up) DOS(down) PDOS(up) PDOS(down)
            if self.k_resolved:
                reshaped_pdos = dos[4:6].reshape((2, self.nkpoints, self.nepoints))[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
                # Squeeze if necessary
                if squeeze:
                    return np.sum(reshaped_pdos, axis=1) / self.nkpoints
                return reshaped_pdos

            # E DOS(up) DOS(down) PDOS(up) PDOS(down)
            return dos[3:5][:, self.energy_window[0] : self.energy_window[1]]

        elif self.case == 3:
            # ik E DOS PDOS(up) PDOS(down)
            if self.k_resolved:
                reshaped_pdos = dos[3:5].reshape((2, self.nkpoints, self.nepoints))[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
                # Squeeze if necessary
                if squeeze:
                    total_pdos = np.sum(reshaped_pdos, axis=1) / self.nkpoints
                return reshaped_pdos

            # E DOS PDOS(up) PDOS(down)
            return dos[2:4][:, self.energy_window[0] : self.energy_window[1]]

        # ik E DOS PDOS
        if self.k_resolved:
            reshaped_pdos = dos[3].reshape((self.nkpoints, self.nepoints))[
                :, self.energy_window[0] : self.energy_window[1]
            ]
            # Squeeze if necessary
            if squeeze:
                return np.sum(reshaped_pdos, axis=0) / self.nkpoints
            return reshaped_pdos

        # E DOS PDOS
        return dos[2][self.energy_window[0] : self.energy_window[1]]

    @property
    def atoms(self):
        r"""
        List of atom's types.

        Returns
        -------
        atoms : list
            List of atom's types symbols

            .. code-block:: python

                atoms = [atom_symbol, ...]
        """

        return [atom for atom in self._atoms]

    def atom_numbers(self, atom: str):
        r"""
        List of atom's numbers for particular atom type.

        Parameters
        ----------
        atom : str
            Label of atom.

        Returns
        -------
        numbers : list

            .. code-block:: python

                numbers = [atom_number, ...]
        """

        return self._atoms[atom]

    def wfcs(self, atom: str, atom_number: int = None):
        r"""
        Return list of wave function symbols for particular atom.

        Parameters
        ----------
        atom : str
            Label of an atom.
        atom_number : int, optional
            Number of an atom. If ``None`` then return wfc symbols of first atom.

        Returns
        -------
        wfc_symbols : list
            List of wave function symbols

            .. code-block:: python

                wfc_symbols = [wfc_symbol, ...]

        """

        # If atom_number is not provided, then return wfc symbols of first atom
        if atom_number is None:
            atom_number = self.atom_numbers(atom)[0]
        return [symbol for symbol in self._wfcs[(atom, atom_number)]]

    def wfc_numbers(self, atom: str, wfc_symbol, atom_number: int = None):
        r"""
        Return list of wave function numbers for particular atom and wave function type.

        Parameters
        ----------
        atom : str
            Label of an atom.
        wfc_symbol : str
            Label of a wave function type.
        atom_number : int, optional
            Number of an atom. If ``None`` then return wfc numbers of first atom.

        Returns
        -------
        wfc_numbers : list
            List of wave function numbers

            .. code-block:: python

                wfc_numbers = [wfc_number, ...]

        """

        # If atom_number is not provided, then return wfc numbers of first atom
        if atom_number is None:
            atom_number = self.atom_numbers(atom)[0]
        return self._wfcs[(atom, atom_number)][wfc_symbol]

    def pdos(
        self, atom, wfc, wfc_numbers=None, atom_numbers=None, background_total=False
    ) -> PDOSQE:
        r"""
        Projected density of states for a particular atom.

        Parameters
        ----------
        atom : str
            Name of the atom type.
        wfc : str
            Name of the projector wave function.
        wfc_numbers : Iterable or int, optional
            Number of wave function projector.
        atom_numbers : Iterable or int, optional
            If ``None``, then PDOS summed over all atom numbers for ``atom``.

        Returns
        -------
        pdos : :py:class:`PDOSQE`
            Partial density of states, orbital-resolved.
        """

        # Check if atom is valid
        if atom not in self.atoms:
            raise ValueError(f"There is no {atom} in PDOS.")

        # Prepare atom_numbers
        if atom_numbers is None:
            atom_numbers = self.atom_numbers(atom)
        elif isinstance(atom_numbers, int):
            atom_numbers = [atom_numbers]
        elif not isinstance(atom_numbers, Iterable):
            raise ValueError(
                f"atom_numbers must be None, int or Iterable, "
                + f"got {type(atom_numbers)} :\n"
                + f"{atom_numbers}"
            )

        # Check if wfc is valid
        if wfc not in self.wfcs(atom, atom_numbers[0]):
            raise ValueError(f"There is no {wfc} projector of {atom} atom in PDOS.")

        # Load and sum data for all atom numbers of the atom and all wfc numbers
        for i, atom_number in enumerate(atom_numbers):
            # Check if atom number is valid
            if atom_number not in self.atom_numbers(atom):
                raise ValueError(f"There is no {atom}#{atom_number} atom in PDOS.")

            # Prepare wfc_numbers
            if wfc_numbers is None:
                wfc_numbers = self.wfc_numbers(atom, wfc, atom_number)
            elif isinstance(wfc_numbers, int):
                wfc_numbers = [wfc_numbers]
            elif not isinstance(wfc_numbers, Iterable):
                raise ValueError(
                    f"wfc_numbers must be None, int or Iterable,"
                    + f" got {type(wfc_numbers)} :\n"
                    + f"{wfc_numbers}"
                )

            for j, wfc_number in enumerate(wfc_numbers):
                # Check if wfc number is valid
                if wfc_number not in self.wfc_numbers(atom, wfc, atom_number):
                    raise ValueError(
                        f"There is no {wfc}#{wfc_number} wave-function for "
                        + f"{atom}#{atom_number} atom in PDOS."
                    )
                # Load and sum data
                path = join(
                    self._input_folder,
                    f"{self.seedname}.pdos_atm#{atom_number}({atom})_wfc#{wfc_number}({wfc})",
                )
                if i == 0 and j == 0:
                    pdos = np.loadtxt(path, skiprows=1).T
                else:
                    pdos += np.loadtxt(path, skiprows=1).T

        # Reshape data
        if self.spin_pol:
            if self.k_resolved:
                # ik E LDOS(up) LDOS(down) PDOS_1(up) PDOS_1(down) ... PDOS_2l+1(up) PDOS_2l+1(down)
                ldos = pdos[2:4].reshape(2, self.nkpoints, self.nepoints)[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
                pdos = pdos[4:].reshape(
                    (pdos.shape[0] - 4) // 2, 2, self.nkpoints, self.nepoints
                )[:, :, :, self.energy_window[0] : self.energy_window[1]]
            else:
                # E LDOS(up) LDOS(down) PDOS_1(up) PDOS_1(down) ... PDOS_2l+1(up) PDOS_2l+1(down)
                ldos = pdos[1:3][:, self.energy_window[0] : self.energy_window[1]]
                pdos = pdos[3:].reshape((pdos.shape[0] - 3) // 2, 2, self.nepoints)[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
        else:
            if self.k_resolved:
                # ik E LDOS PDOS_1 ... PDOS_2l+1
                ldos = pdos[2].reshape(self.nkpoints, self.nepoints)[
                    :, self.energy_window[0] : self.energy_window[1]
                ]
                pdos = pdos[3:].reshape(
                    pdos.shape[0] - 3, self.nkpoints, self.nepoints
                )[:, :, self.energy_window[0] : self.energy_window[1]]
            else:
                # E LDOS PDOS_1 ... PDOS_2l+1
                ldos = pdos[1][self.energy_window[0] : self.energy_window[1]]
                pdos = pdos[2:][:, self.energy_window[0] : self.energy_window[1]]

        if background_total:
            ldos = self.total_pdos()
        return PDOSQE(
            energy=self.energy,
            pdos=pdos,
            projectors_group=wfc,
            ldos=ldos,
            spin_pol=self.spin_pol,
        )

    def plot_pdos_tot(
        self,
        output_name,
        interactive=False,
        efermi=0,
        xlim=None,
        ylim=None,
        save_pickle=False,
        axes_labels_fontsize=18,
        legend_fontsize=12,
        title_fontsize=18,
    ):
        r"""
        Plot total DOS vs total PDOS.

        Parameters
        ----------
        output_name : str
            Name of the output file.
        interactive : bool, default False
            Whether to plot in interactive |matplotlib|_ window.
        efermi : float, default 0
            Fermi energy. Zero of energy scale is shifted to it.
        xlim : tuple, optional
            Limits for energy scale.
        ylim : tuple, optional
            Limits for dos scale.
        save_pickle : bool, default False
            Whether to save figure as a .pickle file.
            Helps for custom modification of particular figures.
        axes_label_fontsize : int, default 18
            Fontsize of the axes labels.
        legend_fontsize : int, default 12
            Fontsize of the legend.
        title_fontsize : int, default 18
            Title fontsize.
        """
        fig, ax = plt.subplots(figsize=(8, 4))

        #  x axis label
        if efermi == 0:
            ax.set_xlabel("E, eV", fontsize=axes_labels_fontsize)
        else:
            ax.set_xlabel("E - E$_{Fermi}$, eV", fontsize=axes_labels_fontsize)
        # y axis label
        ax.set_ylabel("DOS, states/eV", fontsize=axes_labels_fontsize)

        # x axis limits
        if xlim is None:
            xlim = (np.amin(self.energy), np.amax(self.energy))
        ax.set_xlim(*tuple(xlim))
        # y axis limits
        if ylim is not None:
            ax.set_ylim(*tuple(ylim))

        # Title
        ax.set_title(f"DOS vs PDOS", fontsize=title_fontsize)

        # Eye-guide lines
        if efermi != 0:
            plot_vlines(ax, 0)
        if self.spin_pol:
            plot_hlines(ax, 0)

        def fill(data, color, label):
            ax.fill_between(
                self.energy, 0, data, lw=0, color=color, alpha=0.3, label=label
            )

        def plot(data, color, label):
            ax.plot(self.energy, data, "-", lw=1, color=color, alpha=0.7, label=label)

        # E DOS(E) PDOS(E)
        if self.case in [1, 4]:
            fill(self.total_dos(squeeze=True), "grey", "DOS")
            plot(self.total_pdos(squeeze=True), "black", "PDOS")
            ncol = 1
        # E DOSup(E) DOSdw(E) PDOSup(E) PDOSdw(E)
        if self.case == 2:
            fill(self.total_dos(squeeze=True)[0], "blue", "DOS (up)")
            fill(-self.total_dos(squeeze=True)[1], "red", "DOS (down)")
            plot(self.total_pdos(squeeze=True)[0], "blue", "PDOS (up)")
            plot(-self.total_pdos(squeeze=True)[1], "red", "PDOS (down)")
            ncol = 2
        # E DOS(E) PDOSup(E) PDOSdw(E)
        if self.case == 3:
            fill(self.total_dos(squeeze=True), "grey", "DOS")
            fill(-self.total_dos(squeeze=True), "grey", "$-$DOS")
            plot(self.total_pdos(squeeze=True)[0], "blue", "PDOS (up)")
            plot(-self.total_pdos(squeeze=True)[1], "red", "PDOS (down)")
            ncol = 1

        if interactive:
            ax.legend(loc="best", ncol=ncol, draggable=True, fontsize=legend_fontsize)
        else:
            ax.legend(loc="best", ncol=ncol, fontsize=legend_fontsize)

        if interactive:
            plt.show()
        else:
            png_path = f"{output_name}.png"
            plt.savefig(png_path, dpi=600, bbox_inches="tight")
            if save_pickle:
                import pickle

                with open(f"{png_path}.pickle", "wb") as file:
                    pickle.dump(fig, file)
        plt.close()


class DOSIterator:
    def __init__(self, dos: DOSQE) -> None:
        self._list = []
        for atom in dos.atoms:
            for atom_number in dos.atom_numbers(atom):
                for wfc in dos.wfcs(atom, atom_number):
                    for wfc_number in dos.wfc_numbers(atom, wfc, atom_number):
                        self._list.append([atom, atom_number, wfc, wfc_number])
        self._index = 0

    def __next__(self) -> str:
        if self._index < len(self._list):
            result = self._list[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self):
        return self


def detect_seednames(input_folder):
    r"""
    Analyze input folder, detects seednames for the dos output files.

    Parameters
    ----------
    input_folder : str
        Directory with DOS files.

    Returns
    -------
    seednames : list
        List of seednames found in ``input_folder``.
    """

    # Get list of files in the folder
    files = []
    for _, _, filenames in walk(input_folder):
        files.extend(filenames)
        break

    seednames = set()
    for file in files:
        if ".pdos_tot" in file and ".pdos_tot" == file[-9:]:
            seednames.add(file[:-9])
        elif re.match(f".*{PATTERN}$", file):
            seednames.add(re.split(f"{PATTERN}$", file)[0])
    seednames = list(seednames)

    return seednames


def prepare_custom_pdos(dos: DOSQE, custom, quiet=True):
    r"""
    Prepare custom PDOS. Based on the input line.

    Parameters
    ----------
    dos : :py:class:`.DOSQE`
        DOS input files wrapper.
    custom : list of str
        List of strings describing the custom PDOS.
        See :ref:`rad-plot-dos_custom-plots` for the string`s description.
    quiet : bool, default True
        Whether to print information about the runtime.

    Returns
    -------
    pdos : list of :numpy:`ndarray`
        PDOS array. Not the instance of the :py:class:`.PDOS`, but
        an array, which could be passed to the constructor of
        :py:class:`.PDOS` class
    """
    pdos = []

    # Process each custom specification
    for entry in custom:
        # Print info about current specification
        if not quiet:
            cprint(f'"{entry}":', "green")

        # Initialize pdos_element corresponding to the entry
        pdos_element = None

        # Remove spaces and closing brackets
        entry = entry.replace(" ", "").replace(")", "")

        # Split by semicolon into different atom types
        subentries = entry.split(";")

        # Process each atom specification
        for subentry in subentries:
            # Get the info about atoms
            atom_part = subentry.split("(")[0]

            # Get the info about atom type
            atom = atom_part.split("#")[0]

            # Get the info about atom numbers
            if "#" in atom_part:
                # Atom numbers provided directly
                atom_numbers = list(map(int, atom_part.split("#")[1:]))
                # Check if the custom numbers are correct
                if not set(atom_numbers).issubset(set(dos.atom_numbers(atom))):
                    raise ValueError(
                        f"For the atom {atom} the following numbers exist:\n"
                        + f"{dos.atom_numbers(atom)}\n"
                        + f"but the following numbers are provided:\n"
                        + f"{atom_numbers}"
                    )
            else:
                # Atom numbers are taken from the dos object
                atom_numbers = dos.atom_numbers(atom)

            # Print the summary of the atom with numbers
            if not quiet:
                cprint(
                    f"  * PDOS is summed among the following {atom} atoms:",
                    "green",
                    end="\n      ",
                )
                for i, number in enumerate(atom_numbers):
                    print(f"{atom}#{number}", end="")
                    if i != len(atom_numbers) - 1:
                        print(end=", ")
                    else:
                        print()

            # Get the info about wave function projectors, if any
            if "(" in subentry:
                # Get all wave function types
                wfc_parts = subentry.split("(")[1].split(",")
                wfcs = {}
                # Process each wave function type
                for wfc_part in wfc_parts:
                    # Get the name of the wave function type
                    wfc = wfc_part.split("#")[0]

                    # If wave function numbers are provided directly
                    if "#" in wfc_part:
                        wfc_numbers = list(map(int, wfc_part.split("#")[1:]))
                        if wfc in wfcs:
                            wfcs[wfc].extend(wfc_numbers)
                        else:
                            wfcs[wfc] = wfc_numbers
                    # If wave function numbers are not provided
                    else:
                        # It will result in the summ over all numbers
                        wfcs[wfc] = dos.wfc_numbers(atom, wfc, atom_numbers[0])

            # If no wave function projectors are provided, then sum over all projectors
            else:
                wfcs = dict(
                    [
                        (wfc, dos.wfc_numbers(atom, wfc, atom_numbers[0]))
                        for wfc in dos.wfcs(atom)
                    ]
                )

            if not quiet:
                print(
                    f"  * For each {atom} atom PDOS is summed among the following projections:",
                    end="\n      ",
                )
                for wfc in wfcs:
                    for number in wfcs[wfc]:
                        print(f"{wfc}#{number}", end=" ")
                print()

            for wfc in wfcs:
                tmp_pdos = dos.pdos(
                    atom=atom,
                    wfc=wfc,
                    wfc_numbers=wfcs[wfc],
                    atom_numbers=atom_numbers,
                ).ldos
                if pdos_element is None:
                    pdos_element = tmp_pdos
                else:
                    pdos_element += tmp_pdos

        pdos.append(pdos_element)

    return np.array(pdos)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "PREDEFINED_NOTATIONS",
    "TXT_FLAGS",
]

PREDEFINED_NOTATIONS = {
    "standard": (True, False, -1),
    "tb2j": (True, True, -1),
    "spinw": (True, False, 1),
    "vampire": (True, True, -0.5),
}

TXT_FLAGS = {
    "cell": "Primitive unit cell:",
    "atoms": "Atoms:",
    "magmoms": "Magnetic moments of magnetic atoms:",
    "spins": "Spin of magnetic atoms:",
    "spinham": "Spin Hamiltonian:",
    "iso": "Isotropic exchange:",
    "matrix": "Full exchange matrix:",
    "dmi": "DMI vector:",
    "dmis": "DMI vectors:",
    "dmi_module": "DMI module:",
    "dmi_relative": "|DMI| / |J_iso|:",
    "aniso": "Symmetric anisotropic exchange:",
    "spinmodel": "Spin model:",
    "kpath": "K-path:",
    "kpoints": "Points (in relative coordinates):",
    "klabels": "Labels:",
    "dispersion": "Magnon dispersion:",
    "separate": "Data are in the file:",
    "notation": "Notation of Hamiltonian:",
    "bonds": "Bonds:",
}



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["ColpaFailed", "NotationError"]


class ColpaFailed(Exception):
    r"""
    Raised when Diagonalization via Colpa fails.
    """

    def __init__(self):
        self.message = "Diagonalization via Colpa failed."

    def __str__(self):
        return self.message


class NotationError(ValueError):
    r"""
    Raised when the notation (or individual property) is not defined.

    Gives a summary of the notation (or individual property) and how to set it.

    Parameters
    ----------
    name : str
        Name of the corresponding attribute.

    """

    def __init__(self, name):
        self.message = (
            f"\n\nNotation`s interpretation is not set for the property {name}.\n"
            + f"Set the notation first:\n"
            + f"    SpinHamiltonian.{name} = True  "
            + f"or  SpinHamiltonian.{name} = False\n\n"
            + f"Note: When the attribute is set for the first time it sets the interpretation, "
            + "afterwards it change the notation.\n\n"
            + f"If you want to set the interpretation again, use \n"
            + f"    SpinHamiltonian.set_interpretation({name} = True)"
            + "\nor\n"
            + f"    SpinHamiltonian.set_interpretation({name} = False)\n"
        )

    def __str__(self):
        return self.message



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Script interface to the radtools package.

The behaviour of the command line interface is defined by the
functions in this module.

The functions are called with the same names as the scripts,
but the prefix "rad-" is removed and "-" are substituted by "_".
Function`s arguments directly correspond to the full names of the
arguments of the script (i.e. the argument :ref:`rad-extract-tb2j_input-filename`
of the script :ref:`rad-extract-tb2j` is passed to the function :py:func:`.extract_tb2j` as the
argument ``input_filename``).

Full documentation on the behaviour is available in the
:ref:`scripts-guide`.


.. admonition:: Example

    Identification of Wannier centres from the file "seedname_centres.xyz"
    with increased span of 0.2, saving the result in the file
    "identified_centres" of the current directory:

    .. code-block:: bash

        rad-identify-wannier-centres.py seedname_centres.xyz -s 0.2 -on identified_centres

    The same result could be achieved by calling the function
    :py:func:`.identify_wannier_centres`:

    .. code-block:: python

        from radtools import identify_wannier_centres
        identify_wannier_centres("seedname_centres.xyz",
            span = 0.2,
            output_name="identified_centres")
"""

from .extract_tb2j import manager as extract_tb2j
from .identify_wannier_centres import manager as identify_wannier_centres
from .make_template import manager as make_template
from .plot_dos import manager as plot_dos
from .plot_fatbands import manager as plot_fatbands
from .plot_tb2j import manager as plot_tb2j
from .plot_tb2j_magnons import manager as plot_tb2j_magnons

__all__ = [
    "identify_wannier_centres",
    "plot_tb2j",
    "extract_tb2j",
    "plot_dos",
    "make_template",
    "plot_tb2j_magnons",
    "plot_fatbands",
]
